from rich.text import Text
from textual import events, on
from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import DataTable, Input
from textual.widgets._data_table import ColumnKey

from bibtui.bib.models import READ_STATES, BibEntry
from bibtui.pdf.paths import find_pdf_for_entry
from bibtui.utils.dates import extract_date_added, format_bib_date

# Original header labels in column order
_COL_LABELS = (
    "◉",
    "!",
    "◫",
    "🔗",
    "Type",
    "Year",
    "Author",
    "Journal",
    "Title",
    "Added",
    "★",
)

_COL_OVERHEAD_WITH_JOURNAL_AND_ADDED = 85
_COL_OVERHEAD_WITH_JOURNAL_NO_ADDED = 75
_COL_OVERHEAD_NO_JOURNAL_WITH_ADDED = 68
_COL_OVERHEAD_NO_JOURNAL_NO_ADDED = 58
_JOURNAL_THRESHOLD = 120  # min widget width (chars) to show the Journal column
_ADDED_THRESHOLD = 140  # min widget width (chars) to show the Date Added column

_FIELD_PREFIXES: dict[str, str] = {
    "t": "title",
    "title": "title",
    "a": "author",
    "author": "author",
    "j": "journal",
    "journal": "journal",
    "k": "keywords",
    "kw": "keywords",
    "keyword": "keywords",
    "keywords": "keywords",
    "y": "year",
    "year": "year",
    "u": "url",
    "url": "url",
    "c": "citekey",
    "citekey": "citekey",
    "key": "citekey",
}


def _parse_query(query: str) -> tuple[list[tuple[str, str]], list[str]]:
    """Split a query into field filters and free-text terms.

    Each space-separated token is either ``prefix:value`` (field filter) or a
    plain word (searched across all fields).  Multiple tokens are ANDed.
    The keyword ``AND`` (case-insensitive) is ignored, allowing queries like
    ``j:nature AND y:2025``.
    """
    filters: list[tuple[str, str]] = []
    free_terms: list[str] = []
    for token in query.split():
        if token.upper() == "AND":
            continue
        if ":" in token:
            prefix, _, value = token.partition(":")
            field = _FIELD_PREFIXES.get(prefix.lower())
            if field and value:
                filters.append((field, value.lower()))
                continue
        free_terms.append(token.lower())
    return filters, free_terms


def _entry_matches(
    entry, filters: list[tuple[str, str]], free_terms: list[str]
) -> bool:
    for field, value in filters:
        if field == "title":
            if value not in entry.title.lower():
                return False
        elif field == "author":
            if value not in entry.author.lower():
                return False
        elif field == "keywords":
            if value not in entry.keywords.lower():
                return False
        elif field == "year":
            if "-" in value:
                # Range: y:2010-2020
                parts = value.split("-", 1)
                try:
                    y_min, y_max = int(parts[0]), int(parts[1])
                    y = int(entry.year) if entry.year.isdigit() else 0
                    if not (y_min <= y <= y_max):
                        return False
                except ValueError:
                    if value not in entry.year:
                        return False
            else:
                if value not in entry.year:
                    return False
        elif field == "journal":
            journal = (entry.journal or entry.raw_fields.get("booktitle", "")).lower()
            if value not in journal:
                return False
        elif field == "url":
            if value not in entry.url.lower():
                return False
        elif field == "citekey":
            if value not in entry.key.lower():
                return False
    for term in free_terms:
        if not (
            term in entry.title.lower()
            or term in entry.author.lower()
            or term in entry.keywords.lower()
            or term in entry.key.lower()
        ):
            return False
    return True


class EntryList(Widget):
    """Left pane: searchable DataTable of BibTeX entries."""

    ALLOW_MAXIMIZE = True

    DEFAULT_CSS = """
    EntryList {
        layout: vertical;
        height: 100%;
    }
    EntryList Input {
        height: 3;
    }
    EntryList DataTable {
        height: 1fr;
    }
    """

    BORDER_TITLE = "Entries"

    search_text: reactive[str] = reactive("")

    def __init__(self, entries: list[BibEntry], **kwargs):
        super().__init__(**kwargs)
        self._all_entries: list[BibEntry] = entries
        self._filtered: list[BibEntry] = list(entries)
        self._col_keys: tuple[ColumnKey, ...] = ()
        self._col_state: ColumnKey | None = None
        self._col_priority: ColumnKey | None = None
        self._col_file: ColumnKey | None = None
        self._col_url: ColumnKey | None = None
        self._col_journal: ColumnKey | None = None
        self._col_title: ColumnKey | None = None
        self._col_added: ColumnKey | None = None
        self._col_rating: ColumnKey | None = None
        self._title_width: int = 30
        self._journal_visible: bool = True
        self._added_visible: bool = True
        self._sort_key: ColumnKey | None = None
        self._sort_reverse: bool = False
        self._pdf_base_dir: str = ""

    def set_pdf_base_dir(self, base_dir: str) -> None:
        self._pdf_base_dir = base_dir

    def _file_icon(self, entry: BibEntry) -> str:
        if not entry.file:
            return " "
        return (
            "■"
            if find_pdf_for_entry(entry.file, entry.key, self._pdf_base_dir)
            else "□"
        )

    def compose(self) -> ComposeResult:
        yield Input(
            placeholder="Search… (a:smith j:nature y:2025 k:ice c:smith2020)",
            id="search-input",
        )
        yield DataTable(id="entry-table", cursor_type="row")

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        col_state = table.add_column("◉", width=1)
        col_priority = table.add_column("!", width=1)
        col_file = table.add_column("◫", width=1)
        col_url = table.add_column("🔗", width=2)
        col_type = table.add_column("Type", width=7)
        col_year = table.add_column("Year", width=4)
        col_author = table.add_column("Author", width=13)
        col_journal = table.add_column("Journal", width=17)
        col_title = table.add_column("Title", width=self._title_width)
        col_added = table.add_column("Added", width=10)
        col_rating = table.add_column("★", width=5)
        self._col_keys = (
            col_state,
            col_priority,
            col_file,
            col_url,
            col_type,
            col_year,
            col_author,
            col_journal,
            col_title,
            col_added,
            col_rating,
        )
        self._col_state = col_state
        self._col_priority = col_priority
        self._col_file = col_file
        self._col_url = col_url
        self._col_journal = col_journal
        self._col_title = col_title
        self._col_added = col_added
        self._col_rating = col_rating
        self._populate_table(self._all_entries)
        self._update_title_width()

    def on_resize(self, event) -> None:
        self._update_title_width()

    def _update_title_width(self) -> None:
        """Recompute title column width to fill available horizontal space.

        Also shows/hides Journal and Added columns based on width thresholds.
        """
        if self._col_title is None:
            return
        show_journal = self.size.width >= _JOURNAL_THRESHOLD
        show_added = self.size.width >= _ADDED_THRESHOLD
        table = self.query_one(DataTable)
        if self._col_journal is not None and show_journal != self._journal_visible:
            self._journal_visible = show_journal
            table.columns[self._col_journal].width = 17 if show_journal else 0
        if self._col_added is not None and show_added != self._added_visible:
            self._added_visible = show_added
            table.columns[self._col_added].width = 10 if show_added else 0
        if self._journal_visible and self._added_visible:
            overhead = _COL_OVERHEAD_WITH_JOURNAL_AND_ADDED
        elif self._journal_visible and not self._added_visible:
            overhead = _COL_OVERHEAD_WITH_JOURNAL_NO_ADDED
        elif not self._journal_visible and self._added_visible:
            overhead = _COL_OVERHEAD_NO_JOURNAL_WITH_ADDED
        else:
            overhead = _COL_OVERHEAD_NO_JOURNAL_NO_ADDED
        width = max(10, self.size.width - overhead)
        if width == self._title_width:
            return
        self._title_width = width
        table.columns[self._col_title].width = width
        table.refresh()

    @staticmethod
    def _date_added_text(entry: BibEntry) -> str:
        return format_bib_date(extract_date_added(entry.raw_fields))

    def _populate_table(self, entries: list[BibEntry]) -> None:
        table = self.query_one(DataTable)
        table.clear()
        self._filtered = entries
        for e in entries:
            journal = e.journal or e.raw_fields.get("booktitle", "")
            table.add_row(
                e.read_state_icon,
                e.priority_icon,
                self._file_icon(e),
                e.url_icon,
                e.entry_type[:7],
                e.year[:4] if e.year else "",
                e.authors_short[:12] + "…"
                if len(e.authors_short) > 12
                else e.authors_short,
                journal[:16] + "…" if len(journal) > 16 else journal,
                e.title,
                self._date_added_text(e),
                e.rating_stars,
                key=e.key,
            )

    # ── Sorting ───────────────────────────────────────────────────────────

    @on(DataTable.HeaderSelected)
    def on_header_selected(self, event: DataTable.HeaderSelected) -> None:
        col_key = event.column_key
        if self._sort_key == col_key:
            self._sort_reverse = not self._sort_reverse
        else:
            self._sort_key = col_key
            self._sort_reverse = False
        self._apply_sort()
        self._update_header_labels()

    def _sort_fn(self, col_key: ColumnKey):
        """Return a key function for sorting BibEntry by the given column."""
        idx = list(self._col_keys).index(col_key)
        if idx == 0:  # ◉ read state
            return lambda e: (
                READ_STATES.index(e.read_state) if e.read_state in READ_STATES else 0
            )
        if idx == 1:  # ! priority
            return lambda e: e.priority if e.priority > 0 else 99
        if idx == 2:  # ◫ file
            return lambda e: 0 if e.file else 1
        if idx == 3:  # ⊕ url
            return lambda e: 0 if e.url else 1
        if idx == 4:  # Type
            return lambda e: e.entry_type
        if idx == 5:  # Year
            return lambda e: int(e.year) if e.year.isdigit() else 0
        if idx == 6:  # Author
            return lambda e: e.authors_short.lower()
        if idx == 7:  # Journal
            return lambda e: (e.journal or e.raw_fields.get("booktitle", "")).lower()
        if idx == 8:  # Title
            return lambda e: e.title.lower()
        if idx == 9:  # Added
            return lambda e: self._date_added_text(e)
        if idx == 10:  # ★ rating
            return lambda e: e.rating
        return lambda e: ""

    def _apply_sort(self) -> None:
        if self._sort_key is None:
            return
        self._filtered = sorted(
            self._filtered,
            key=self._sort_fn(self._sort_key),
            reverse=self._sort_reverse,
        )
        # Rebuild the table rows in new order without re-fetching data
        table = self.query_one(DataTable)
        table.clear()
        for e in self._filtered:
            journal = e.journal or e.raw_fields.get("booktitle", "")
            table.add_row(
                e.read_state_icon,
                e.priority_icon,
                self._file_icon(e),
                e.url_icon,
                e.entry_type[:7],
                e.year[:4] if e.year else "",
                e.authors_short[:12] + "…"
                if len(e.authors_short) > 12
                else e.authors_short,
                journal[:16] + "…" if len(journal) > 16 else journal,
                e.title,
                self._date_added_text(e),
                e.rating_stars,
                key=e.key,
            )

    def _update_header_labels(self) -> None:
        """Put ▲/▼ on the active sort column, restore others."""
        table = self.query_one(DataTable)
        for key, label in zip(self._col_keys, _COL_LABELS):
            if key == self._sort_key:
                indicator = "▼" if self._sort_reverse else "▲"
                table.columns[key].label = Text(f"{label} {indicator}")
            else:
                table.columns[key].label = Text(label)
        table.refresh()

    # ── Search ────────────────────────────────────────────────────────────

    @on(Input.Changed, "#search-input")
    def on_search_changed(self, event: Input.Changed) -> None:
        query = event.value.strip()
        if not query:
            base = self._all_entries
        else:
            filters, free_terms = _parse_query(query)
            base = [
                e for e in self._all_entries if _entry_matches(e, filters, free_terms)
            ]
        self._populate_table(base)
        if self._sort_key is not None:
            self._apply_sort()

    @on(Input.Submitted, "#search-input")
    def on_search_submitted(self, event: Input.Submitted) -> None:
        """Enter in search bar moves focus to the table."""
        self.query_one(DataTable).focus()

    def on_key(self, event: events.Key) -> None:
        """Allow arrow keys to move the table cursor while search is focused."""
        table = self.query_one(DataTable)
        search = self.query_one(Input)
        if self.app.focused is search:
            if event.key == "down":
                table.action_cursor_down()
                event.stop()
            elif event.key == "up":
                table.action_cursor_up()
                event.stop()

    # ── Public helpers ────────────────────────────────────────────────────

    def refresh_entries(self, entries: list[BibEntry]) -> None:
        """Reload all entries (e.g. after add/edit)."""
        table = self.query_one(DataTable)
        selected_before = self.selected_entry
        selected_key = selected_before.key if selected_before is not None else None
        self._all_entries = entries
        search = self.query_one(Input).value
        if search:
            self.on_search_changed(Input.Changed(self.query_one(Input), search))
        else:
            self._populate_table(entries)
            if self._sort_key is not None:
                self._apply_sort()

        if selected_key is None:
            return
        try:
            row_idx = next(
                i for i, e in enumerate(self._filtered) if e.key == selected_key
            )
        except StopIteration:
            return
        table.move_cursor(row=row_idx)

    def refresh_row(self, entry: BibEntry) -> None:
        """Update the read-state, priority, file, and rating cells for a single row."""
        table = self.query_one(DataTable)
        table.update_cell(
            entry.key, self._col_state, entry.read_state_icon, update_width=False
        )
        table.update_cell(
            entry.key, self._col_priority, entry.priority_icon, update_width=False
        )
        table.update_cell(
            entry.key, self._col_file, self._file_icon(entry), update_width=False
        )
        table.update_cell(entry.key, self._col_url, entry.url_icon, update_width=False)
        table.update_cell(
            entry.key, self._col_rating, entry.rating_stars, update_width=False
        )

    @property
    def selected_entry(self) -> BibEntry | None:
        table = self.query_one(DataTable)
        if table.cursor_row < 0 or not self._filtered:
            return None
        if table.cursor_row >= len(self._filtered):
            return None
        return self._filtered[table.cursor_row]

    @property
    def filtered_entries(self) -> list[BibEntry]:
        return self._filtered



import os


def parse_jabref_path(file_field: str, base_dir: str = "") -> str:
    """Resolve a JabRef-style file field to an absolute path.

    JabRef format: ``description:path:type``  (e.g. ``:Smith2023.pdf:PDF``)
    The description and type parts are optional.
    """
    path = file_field.strip()
    if ":" in path:
        parts = path.split(":")
        # ':path:type' → parts = ['', 'path', 'type']
        # 'desc:path:type' → parts = ['desc', 'path', 'type']
        path = parts[1] if len(parts) >= 2 else parts[0]
    path = path.strip()
    if base_dir and not os.path.isabs(path):
        path = os.path.join(base_dir, path)
    return path


def find_pdf_for_entry(
    file_field: str, entry_key: str, base_dir: str = ""
) -> str | None:
    """Return an existing PDF path for an entry, or None.

    First tries the path stored in *file_field*.  If that doesn't exist,
    falls back to a glob search for ``{entry_key}*.pdf`` in *base_dir* to
    handle filename mismatches between JabRef and bibtui naming conventions.
    """
    import glob as _glob

    if file_field:
        path = parse_jabref_path(file_field, base_dir)
        if os.path.exists(path):
            return path

    if base_dir and entry_key:
        matches = _glob.glob(os.path.join(base_dir, f"{entry_key}*.pdf"))
        if matches:
            return matches[0]

    return None


def format_jabref_path(filepath: str, base_dir: str = "") -> str:
    """Format a path as a JabRef file field value ``:filename.pdf:PDF``.

    If ``base_dir`` is set and ``filepath`` is inside it, store only the
    relative filename so the base directory stays configurable.
    """
    if base_dir:
        try:
            rel = os.path.relpath(filepath, base_dir)
            # relpath gives '..' paths if outside base_dir — keep absolute then
            if not rel.startswith(".."):
                filepath = rel
        except ValueError:
            pass  # different drives on Windows
    name = os.path.basename(filepath) if os.path.sep not in filepath else filepath
    # Use just the basename as the stored path to match JabRef convention
    return f":{name}:PDF"



"""PDF fetching utilities.

Tries three strategies in order:
1. arXiv — free PDF via the arXiv API (DOI 10.48550/arXiv.* or arxiv.org URL)
2. Unpaywall — open-access PDF lookup by DOI (requires email)
3. Direct URL — download if the entry's URL points directly to a PDF

Raises FetchError if none of the strategies succeed.
"""

import json
import os
import re
import shutil
import tempfile
import urllib.request
from pathlib import Path
from urllib.parse import urlparse

from bibtui.bib.models import BibEntry


class FetchError(Exception):
    """Raised when all PDF fetch strategies fail."""


# ---------------------------------------------------------------------------
# Filename helpers
# ---------------------------------------------------------------------------

_UNSAFE_RE = re.compile(r'[\\/:*?"<>|{}]')
_WHITESPACE_RE = re.compile(r"\s+")
_MAX_TITLE_LEN = 80


def pdf_filename(entry: BibEntry) -> str:
    """Return a sanitized filename: ``{key} - {title}.pdf``.

    Characters that are unsafe in filenames are stripped.  The title portion
    is truncated to keep paths reasonable.  Falls back to ``{key}.pdf`` when
    the entry has no title.
    """
    key = entry.key or "unknown"
    title = entry.title.strip() if entry.title else ""
    # Remove LaTeX commands and unsafe chars, then normalise whitespace
    title = _UNSAFE_RE.sub("", title)
    title = _WHITESPACE_RE.sub(" ", title).strip()
    if title:
        if len(title) > _MAX_TITLE_LEN:
            title = title[:_MAX_TITLE_LEN].rstrip()
        return f"{key} - {title}.pdf"
    return f"{key}.pdf"


def add_pdf(src: Path, entry: BibEntry, base_dir: str) -> Path:
    """Move an existing PDF to the canonical location for *entry*.

    Parameters
    ----------
    src:
        Path to the existing PDF file.  ``~`` is expanded and the path is
        resolved to an absolute location.
    entry:
        The BibTeX entry to associate the file with.
    base_dir:
        Destination directory (from ``Config.pdf_base_dir``).  Must not be empty.

    Returns
    -------
    Path
        The destination path the file was moved to.

    Raises
    ------
    FetchError
        If ``base_dir`` is empty, the source file does not exist, the source
        is not a ``.pdf``, or the destination already exists.
    """
    if not base_dir:
        raise FetchError(
            "PDF base directory is not set. "
            "Open Settings (Ctrl+P → Settings) and set a base directory first."
        )
    src = Path(src).expanduser().resolve()
    if not src.exists():
        raise FetchError(f"File not found: {src}")
    if src.suffix.lower() != ".pdf":
        raise FetchError(f"Not a PDF file: {src.name}")
    dest = Path(base_dir) / pdf_filename(entry)
    if dest.exists() and dest.resolve() != src:
        raise FetchError(f"Destination already exists: {dest}")
    dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.move(str(src), dest)
    return dest


# ---------------------------------------------------------------------------
# arXiv helpers
# ---------------------------------------------------------------------------


def _arxiv_id(entry: BibEntry) -> str | None:
    """Extract an arXiv ID from the entry's DOI or URL, or None."""
    # DOI: 10.48550/arXiv.2301.12345  or  10.48550/arxiv.hep-th/9711200
    if entry.doi:
        m = re.search(r"10\.48550/[aA]r[xX]iv\.(.+)$", entry.doi)
        if m:
            return m.group(1)

    # URL: https://arxiv.org/abs/2301.12345  or  /pdf/2301.12345
    if entry.url:
        m = re.search(
            r"arxiv\.org/(?:abs|pdf)/(.+?)(?:v\d+)?(?:\.pdf)?$",
            entry.url,
            re.IGNORECASE,
        )
        if m:
            return m.group(1)

    return None


# ---------------------------------------------------------------------------
# Download helper
# ---------------------------------------------------------------------------


def _download(url: str, dest_path: str, timeout: int = 30) -> None:
    """Stream *url* to *dest_path*.

    Raises FetchError if the response Content-Type is not PDF or the request
    fails.
    """
    req = urllib.request.Request(
        url,
        headers={
            "User-Agent": (
                "Mozilla/5.0 (X11; Linux x86_64; rv:124.0) Gecko/20100101 Firefox/124.0"
            ),
            "Accept": "application/pdf,*/*",
        },
    )
    dest = Path(dest_path)
    dest.parent.mkdir(parents=True, exist_ok=True)
    tmp_path: str | None = None
    try:
        with urllib.request.urlopen(req, timeout=timeout) as resp:
            content_type = resp.headers.get("Content-Type", "")
            first_chunk = resp.read(65536)
            if not first_chunk:
                raise FetchError(f"URL returned empty response: {url}")

            is_pdf_type = "pdf" in content_type.lower()
            is_pdf_magic = first_chunk.startswith(b"%PDF")
            if not is_pdf_type and not is_pdf_magic:
                raise FetchError(
                    f"URL did not return a PDF (Content-Type: {content_type}): {url}"
                )

            fd, tmp_path = tempfile.mkstemp(
                prefix=f".{dest.name}.", suffix=".tmp", dir=str(dest.parent)
            )
            with os.fdopen(fd, "wb") as f:
                f.write(first_chunk)
                while chunk := resp.read(65536):
                    f.write(chunk)
                f.flush()
                os.fsync(f.fileno())

            os.replace(tmp_path, dest_path)
            tmp_path = None
    except FetchError:
        raise
    except Exception as exc:
        raise FetchError(f"Download failed from {url}: {exc}") from exc
    finally:
        if tmp_path and os.path.exists(tmp_path):
            try:
                os.remove(tmp_path)
            except OSError:
                pass


# ---------------------------------------------------------------------------
# Strategy 1 — arXiv
# ---------------------------------------------------------------------------


def _try_arxiv(entry: BibEntry, dest_path: str) -> str | None:
    """Try to fetch the PDF from arXiv.
    Returns None on success, or an error reason string on failure.
    """
    arxiv_id = _arxiv_id(entry)
    if not arxiv_id:
        return "no arXiv ID found"
    pdf_url = f"https://arxiv.org/pdf/{arxiv_id}"
    try:
        _download(pdf_url, dest_path)
        return None
    except FetchError as exc:
        return str(exc)


# ---------------------------------------------------------------------------
# Strategy 2 — Unpaywall
# ---------------------------------------------------------------------------


def _try_unpaywall(entry: BibEntry, dest_path: str, email: str) -> str | None:
    """Try Unpaywall open-access lookup using only direct ``url_for_pdf`` links.

    Returns None on success, or an error reason string on failure.
    If Unpaywall only knows a landing page (url_for_pdf=None), the paper is
    reported as not fetchable via this strategy.
    """
    if not entry.doi:
        return "entry has no DOI"
    if not email:
        return "no email configured in Settings"
    api_url = f"https://api.unpaywall.org/v2/{entry.doi}?email={email}"
    req = urllib.request.Request(
        api_url,
        headers={
            "User-Agent": (
                "Mozilla/5.0 (X11; Linux x86_64; rv:124.0) Gecko/20100101 Firefox/124.0"
            )
        },
    )
    try:
        with urllib.request.urlopen(req, timeout=15) as resp:
            data = json.loads(resp.read().decode("utf-8"))
    except Exception:
        return "Unpaywall lookup failed"

    # Collect all direct PDF URLs Unpaywall knows about
    pdf_candidates: list[str] = []
    best = data.get("best_oa_location") or {}
    if best.get("url_for_pdf"):
        pdf_candidates.append(best["url_for_pdf"])
    for loc in data.get("oa_locations", []):
        u = loc.get("url_for_pdf")
        if u and u not in pdf_candidates:
            pdf_candidates.append(u)

    if not pdf_candidates:
        return "no direct PDF available"

    last_error = ""
    for url in pdf_candidates:
        try:
            _download(url, dest_path)
            return None
        except FetchError as exc:
            last_error = str(exc)
    return last_error


# ---------------------------------------------------------------------------
# Strategy 3 — Direct URL
# ---------------------------------------------------------------------------


def _try_direct_url(entry: BibEntry, dest_path: str) -> str | None:
    """Try to download the entry's URL directly if it looks like a PDF.
    Returns None on success, or an error reason string on failure.
    """
    if not entry.url:
        return "entry has no URL"
    parsed = urlparse(entry.url)
    if parsed.scheme not in ("http", "https"):
        return "URL scheme is not http/https"
    # Quick HEAD request to check Content-Type before consuming bandwidth
    head_req = urllib.request.Request(
        entry.url,
        method="HEAD",
        headers={
            "User-Agent": (
                "Mozilla/5.0 (X11; Linux x86_64; rv:124.0) Gecko/20100101 Firefox/124.0"
            )
        },
    )
    content_type = ""
    try:
        with urllib.request.urlopen(head_req, timeout=10) as resp:
            content_type = resp.headers.get("Content-Type", "")
    except Exception:
        # Some servers reject HEAD; try GET download as fallback.
        content_type = ""

    if content_type and "pdf" not in content_type.lower():
        return f"URL does not serve a PDF (Content-Type: {content_type})"

    try:
        _download(entry.url, dest_path)
        return None
    except FetchError as exc:
        return str(exc)


# ---------------------------------------------------------------------------
# Public entry point
# ---------------------------------------------------------------------------


def fetch_pdf(
    entry: BibEntry,
    dest_dir: str,
    unpaywall_email: str = "",
    overwrite: bool = False,
) -> str:
    """Fetch a PDF for *entry* and save it under *dest_dir*.

    Tries arXiv → Unpaywall → direct URL in order.

    Returns the saved file path on success.
    Raises FetchError (with a human-readable message) if all strategies fail.
    """
    if not dest_dir:
        raise FetchError(
            "PDF base directory is not set. "
            "Open Settings (Ctrl+P → Settings) and set a base directory first."
        )

    try:
        os.makedirs(dest_dir, exist_ok=True)
    except OSError as exc:
        raise FetchError(f"Could not create destination directory: {dest_dir}: {exc}")

    dest_path = os.path.join(dest_dir, pdf_filename(entry))

    if os.path.exists(dest_path) and not overwrite:
        raise FetchError(f"File already exists: {dest_path}")

    reasons: list[str] = []

    reason = _try_arxiv(entry, dest_path)
    if reason is None:
        return dest_path
    reasons.append(f"arXiv: {reason}")

    reason = _try_unpaywall(entry, dest_path, unpaywall_email)
    if reason is None:
        return dest_path
    reasons.append(f"Unpaywall: {reason}")

    reason = _try_direct_url(entry, dest_path)
    if reason is None:
        return dest_path
    reasons.append(f"Direct URL: {reason}")

    raise FetchError("Could not fetch PDF:\n" + "\n".join(f"  • {r}" for r in reasons))



import os

from rich.syntax import Syntax
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import Label, Static, TextArea

from bibtui.bib.models import BibEntry
from bibtui.bib.parser import entry_to_bibtex_str
from bibtui.pdf.paths import parse_jabref_path


def _render_entry(entry: BibEntry, colors: dict[str, str]) -> str:
    """Build a Rich-formatted string for the main body of the detail pane.

    *colors* is a dict with keys: title, key, required, optional, tag_fg,
    tag_bg, warning.  Values are Rich-compatible color strings (hex or names).
    """
    c = colors
    lines: list[str] = []

    # Title
    lines.append(f"[bold {c['title']}]{entry.title or '(no title)'}[/]")
    lines.append("")

    # Entry type badge
    lines.append(
        f"[dim]@{entry.entry_type}[/dim]  [dim]key:[/dim] [{c['key']}]{entry.key}[/]"
    )
    lines.append("")

    # Keywords as badges
    if entry.keywords_list:
        kw_str = " ".join(
            f"[{c['tag_fg']} on {c['tag_bg']}] {k} [/]" for k in entry.keywords_list
        )
        lines.append(f"[bold]Keywords:[/bold]  {kw_str}")
    else:
        lines.append("[bold]Keywords:[/bold]  [dim](none)[/dim]")

    lines.append("")
    lines.append("─" * 50)
    lines.append("")

    # Key fields
    def field_line(label: str, value: str) -> str:
        if value:
            return f"[{c['required']}]{label:<12}[/] {value}"
        else:
            return f"[dim]{label:<12}[/dim] [dim](empty)[/dim]"

    standard_fields = [
        ("Author", "author"),
        ("Year", "year"),
        ("Journal", "journal"),
        ("DOI", "doi"),
        ("URL", "url"),
    ]

    for label, key in standard_fields:
        lines.append(field_line(label, entry.get_field(key)))

    # Raw extra fields
    if entry.raw_fields:
        lines.append("")
        lines.append("[dim]── Other fields ──[/dim]")
        for k, v in entry.raw_fields.items():
            if v:
                lines.append(f"  [dim]{k:<12}[/dim] {v[:80]}")

    # Abstract
    if entry.abstract:
        lines.append("")
        lines.append("[bold]Abstract:[/bold]")
        words = entry.abstract.split()
        current = ""
        for word in words:
            if len(current) + len(word) + 1 > 70:
                lines.append(f"  {current}")
                current = word
            else:
                current = f"{current} {word}".strip()
        if current:
            lines.append(f"  {current}")

    return "\n".join(lines)


def _render_raw(entry: BibEntry) -> Syntax:
    """Render entry as raw BibTeX with syntax highlighting."""
    return Syntax(entry_to_bibtex_str(entry), "bibtex", theme="monokai", word_wrap=True)


class EntryDetail(Widget):
    """Right pane: formatted entry detail view, togglable to raw BibTeX."""

    DEFAULT_CSS = """
    EntryDetail {
        height: 100%;
        overflow-y: scroll;
        padding: 1 2;
    }
    #detail-meta {
        height: auto;
        layout: horizontal;
        margin-bottom: 1;
    }
    #detail-read-state {
        width: auto;
        margin-right: 3;
    }
    #detail-rating {
        width: auto;
        margin-right: 3;
    }
    #detail-priority {
        width: auto;
        margin-right: 3;
    }
    #detail-file {
        width: auto;
    }
    #detail-url {
        width: auto;
        margin-left: 3;
    }
    #detail-content {
        height: auto;
        color: $text;
    }
    #detail-raw {
        display: none;
        height: 1fr;
    }
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._entry: BibEntry | None = None
        self._raw_mode: bool = False
        self._pdf_base_dir: str = ""

    def on_mount(self) -> None:
        self.app.theme_changed_signal.subscribe(self, self._on_theme_changed)

    def _on_theme_changed(self, _theme) -> None:
        if self._entry is not None:
            self._refresh_content()

    def set_pdf_base_dir(self, base_dir: str) -> None:
        self._pdf_base_dir = base_dir

    def compose(self) -> ComposeResult:
        with Horizontal(id="detail-meta"):
            yield Label("", id="detail-read-state")
            yield Label("", id="detail-priority")
            yield Label("", id="detail-rating")
            yield Label("", id="detail-file")
            yield Label("", id="detail-url")
        yield Static("Select an entry to view details.", id="detail-content")
        yield TextArea("", id="detail-raw", read_only=True)

    def show_entry(self, entry: BibEntry | None) -> None:
        self._entry = entry
        self._refresh_content()

    @property
    def raw_mode(self) -> bool:
        return self._raw_mode

    def toggle_view(self) -> None:
        self._raw_mode = not self._raw_mode
        self._refresh_content()
        mode = "raw BibTeX" if self._raw_mode else "formatted"
        self.border_title = f"Entry Detail [{mode}]"

    def _file_icon(self, entry: BibEntry) -> str:
        if not entry.file:
            return " "
        path = parse_jabref_path(entry.file, self._pdf_base_dir)
        return "■" if os.path.exists(path) else "□"

    def _theme_colors(self) -> dict[str, str]:
        """Return Rich color strings derived from the current Textual theme.

        Only used for semantic highlights (title, key, field labels, tags).
        Plain body text uses CSS ``color: $text`` on the Static widget so that
        it automatically follows the theme without any Python involvement.
        """
        tv = self.app.theme_variables
        return {
            "title": tv.get("text-primary", "cyan"),
            "key": tv.get("text-accent", "yellow"),
            "required": tv.get("text-success", "green"),
            "optional": tv.get("text-accent", "blue"),
            "warning": tv.get("text-warning", "yellow"),
            "tag_fg": "white",
            "tag_bg": tv.get("primary", "dark_green"),
        }

    def _refresh_content(self) -> None:
        read_label = self.query_one("#detail-read-state", Label)
        priority_label = self.query_one("#detail-priority", Label)
        rating_label = self.query_one("#detail-rating", Label)
        file_label = self.query_one("#detail-file", Label)
        url_label = self.query_one("#detail-url", Label)
        content = self.query_one("#detail-content", Static)

        if self._entry is None:
            read_label.update("")
            priority_label.update("")
            rating_label.update("")
            file_label.update("")
            url_label.update("")
            content.update("Select an entry to view details.")
            self.query_one("#detail-raw", TextArea).display = False
            content.display = True
            return

        e = self._entry
        colors = self._theme_colors()

        state_label = e.read_state if e.read_state else "unset"
        read_label.update(f"[bold]Read:[/bold] {e.read_state_icon} {state_label}")

        if e.priority:
            priority_label.update(
                f"[bold]Priority:[/bold] {e.priority_icon} {e.priority_label}"
            )
        else:
            priority_label.update("[dim]Priority: —[/dim]")

        stars = e.rating_stars or "[dim]unrated[/dim]"
        rating_label.update(f"[bold]Rating:[/bold] [{colors['warning']}]{stars}[/]")

        icon = self._file_icon(e)
        if not e.file:
            file_label.update("[dim]PDF: —[/dim]")
        elif icon == "■":
            file_label.update("[bold]PDF:[/bold] ■")
        else:
            file_label.update("[bold]PDF:[/bold] [dim]□ not found[/dim]")

        if e.url:
            short = e.url if len(e.url) <= 40 else e.url[:37] + "…"
            url_label.update(f"[bold]🔗 URL:[/bold] {short}")
        else:
            url_label.update("[dim]🔗 URL: —[/dim]")

        raw = self.query_one("#detail-raw", TextArea)
        if self._raw_mode:
            content.display = False
            raw.display = True
            raw.load_text(entry_to_bibtex_str(e))
        else:
            raw.display = False
            content.display = True
            content.update(_render_entry(e, colors))



from dataclasses import dataclass, field

READ_STATES: list[str] = ["", "to-read", "skimmed", "read"]

# Standard Unicode — readable without nerd fonts, pretty with them
READ_STATE_ICONS: dict[str, str] = {
    "": " ",
    "to-read": "○",
    "skimmed": "◑",
    "read": "●",
}

# JabRef priority: prio1=high, prio2=medium, prio3=low
PRIORITIES: list[int] = [0, 1, 2, 3]
PRIORITY_ICONS: dict[int, str] = {0: " ", 1: "↑", 2: "→", 3: "↓"}
PRIORITY_LABELS: dict[int, str] = {0: "unset", 1: "high", 2: "medium", 3: "low"}

ENTRY_TYPES: dict[str, dict[str, list[str]]] = {
    "article": {
        "required": ["author", "title", "journal", "year"],
        "optional": [
            "volume",
            "number",
            "pages",
            "month",
            "doi",
            "abstract",
            "keywords",
        ],
    },
    "book": {
        "required": ["author", "title", "publisher", "year"],
        "optional": [
            "editor",
            "volume",
            "series",
            "address",
            "edition",
            "month",
            "doi",
        ],
    },
    "inproceedings": {
        "required": ["author", "title", "booktitle", "year"],
        "optional": [
            "editor",
            "volume",
            "series",
            "pages",
            "address",
            "month",
            "organization",
            "publisher",
            "doi",
        ],
    },
    "incollection": {
        "required": ["author", "title", "booktitle", "publisher", "year"],
        "optional": [
            "editor",
            "volume",
            "series",
            "chapter",
            "pages",
            "address",
            "edition",
            "month",
            "doi",
        ],
    },
    "phdthesis": {
        "required": ["author", "title", "school", "year"],
        "optional": ["address", "month", "doi"],
    },
    "mastersthesis": {
        "required": ["author", "title", "school", "year"],
        "optional": ["address", "month"],
    },
    "techreport": {
        "required": ["author", "title", "institution", "year"],
        "optional": ["type", "number", "address", "month", "doi"],
    },
    "misc": {
        "required": [],
        "optional": ["author", "title", "year", "month", "url", "doi"],
    },
}


@dataclass
class BibEntry:
    key: str
    entry_type: str
    title: str = ""
    author: str = ""
    year: str = ""
    journal: str = ""
    doi: str = ""
    url: str = ""
    abstract: str = ""
    keywords: str = ""
    comment: str = ""
    rating: int = 0
    read_state: str = ""  # one of READ_STATES
    priority: int = 0  # 0=unset, 1=high, 2=medium, 3=low (JabRef prio1/prio2/prio3)
    file: str = ""
    raw_fields: dict[str, str] = field(default_factory=dict)

    @property
    def url_icon(self) -> str:
        return "🔗" if self.url else "  "

    @property
    def authors_short(self) -> str:
        """Return first author surname or 'Unknown'."""
        if not self.author:
            return "Unknown"
        parts = self.author.split(" and ")
        first = parts[0].strip()
        if "," in first:
            return first.split(",")[0].strip()
        words = first.split()
        return words[-1] if words else first

    @property
    def title_short(self) -> str:
        if len(self.title) <= 60:
            return self.title
        return self.title[:57] + "..."

    @property
    def read_state_icon(self) -> str:
        return READ_STATE_ICONS.get(self.read_state, " ")

    def cycle_read_state(self) -> None:
        idx = (
            READ_STATES.index(self.read_state) if self.read_state in READ_STATES else 0
        )
        self.read_state = READ_STATES[(idx + 1) % len(READ_STATES)]

    @property
    def priority_icon(self) -> str:
        return PRIORITY_ICONS.get(self.priority, " ")

    @property
    def priority_label(self) -> str:
        return PRIORITY_LABELS.get(self.priority, "unset")

    def cycle_priority(self) -> None:
        self.priority = (self.priority + 1) % 4

    @property
    def keywords_list(self) -> list[str]:
        return [k.strip() for k in self.keywords.split(",") if k.strip()]

    @property
    def rating_stars(self) -> str:
        if self.rating == 0:
            return ""
        return "★" * self.rating

    def get_field(self, name: str) -> str:
        if name == "title":
            return self.title
        if name == "author":
            return self.author
        if name == "year":
            return self.year
        if name == "journal":
            return self.journal
        if name == "doi":
            return self.doi
        if name == "url":
            return self.url
        if name == "abstract":
            return self.abstract
        if name == "keywords":
            return self.keywords
        if name == "rating":
            return str(self.rating)
        if name == "file":
            return self.file
        return self.raw_fields.get(name, "")

    def set_field(self, name: str, value: str) -> None:
        if name == "title":
            self.title = value
        elif name == "author":
            self.author = value
        elif name == "year":
            self.year = value
        elif name == "journal":
            self.journal = value
        elif name == "doi":
            self.doi = value
        elif name == "url":
            self.url = value
        elif name == "abstract":
            self.abstract = value
        elif name == "keywords":
            self.keywords = value
        elif name == "rating":
            try:
                self.rating = max(0, min(5, int(value)))
            except ValueError:
                pass
        elif name == "file":
            self.file = value
        else:
            self.raw_fields[name] = value



from textual import events, on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Input,
    Label,
    ListItem,
    ListView,
    LoadingIndicator,
    SelectionList,
    Static,
    Switch,
    TextArea,
)
from textual.widgets._selection_list import Selection

from bibtui.bib.models import BibEntry
from bibtui.bib.parser import bibtex_str_to_entry, entry_to_bibtex_str
from bibtui.utils.config import Config


def _format_age(mtime: float) -> str:
    """Human-readable age string for a file modification time."""
    import time

    age = time.time() - mtime
    if age < 60:
        return "just now"
    if age < 3600:
        return f"{int(age / 60)} min ago"
    if age < 86400:
        return f"{int(age / 3600)} hr ago"
    return f"{int(age / 86400)} days ago"


class ConfirmModal(ModalScreen[bool]):
    """Generic yes/no confirmation dialog."""

    BINDINGS = [Binding("escape", "cancel", "Cancel", show=False)]

    DEFAULT_CSS = """
    ConfirmModal {
        align: center middle;
    }
    ConfirmModal > Vertical {
        width: 50;
        height: auto;
        border: double $warning;
        background: $surface;
        padding: 1 2;
    }
    ConfirmModal #btn-yes {
        background: $error;
        color: $text;
    }
    ConfirmModal #btn-no {
        background: $primary;
        color: $text;
    }
    """

    def __init__(self, message: str, **kwargs):
        super().__init__(**kwargs)
        self._message = message

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Confirm[/bold]", classes="modal-title")
            yield Static(self._message)
            with Horizontal(classes="modal-buttons"):
                yield Button("Yes", id="btn-yes")
                yield Button("No", id="btn-no")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(event.button.id == "btn-yes")

    def action_cancel(self) -> None:
        self.dismiss(False)


class LibraryFetchConfirmModal(ModalScreen[tuple[bool, bool] | None]):
    """Confirm library PDF fetch and choose whether broken links may be overwritten."""

    BINDINGS = [Binding("escape", "cancel", "Cancel", show=False)]

    DEFAULT_CSS = """
    LibraryFetchConfirmModal {
        align: center middle;
    }
    LibraryFetchConfirmModal > Vertical {
        width: 50;
        height: auto;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    LibraryFetchConfirmModal #overwrite-row {
        layout: horizontal;
        align: left middle;
        margin: 1 0;
        height: 3;
    }
    LibraryFetchConfirmModal #overwrite-switch {
        margin-right: 1;
    }
    LibraryFetchConfirmModal #library-fetch-note {
        color: $text-muted;
    }
    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Fetch Missing PDFs[/bold]", classes="modal-title")
            yield Static("Run fetch for entries missing local PDFs?")
            with Horizontal(id="overwrite-row"):
                yield Switch(value=True, id="overwrite-switch")
                yield Label("Overwrite broken links")
            yield Static(
                "Turn off to keep entries with broken file paths untouched.",
                id="library-fetch-note",
            )
            with Horizontal(classes="modal-buttons"):
                yield Button("Start", variant="primary", id="btn-start")
                yield Button("Cancel", id="btn-cancel")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
            return
        overwrite = self.query_one("#overwrite-switch", Switch).value
        self.dismiss((True, overwrite))

    def action_cancel(self) -> None:
        self.dismiss(None)


class DOIModal(ModalScreen[BibEntry | None]):
    """Modal to fetch an entry by DOI."""

    BINDINGS = [
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    DOIModal {
        align: center middle;
    }
    DOIModal > Vertical {
        width: 70;
        height: auto;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    #doi-status.fetching { color: $warning; }
    #doi-status.success  { color: $success; }
    #doi-status.error    { color: $error; }
    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Entry from DOI[/bold]", classes="modal-title")
            yield Input(
                placeholder="Enter DOI (e.g. 10.1038/nature12345)", id="doi-input"
            )
            yield Static("", id="doi-status")
            with Horizontal(classes="modal-buttons"):
                yield Button("Fetch", variant="primary", id="btn-fetch")
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self.call_after_refresh(self.query_one("#doi-input", Input).focus)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-fetch":
            self._do_fetch()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        self._do_fetch()

    def _do_fetch(self) -> None:
        doi = self.query_one("#doi-input", Input).value.strip()
        if not doi:
            return
        status = self.query_one("#doi-status", Static)
        status.set_classes("fetching")
        status.update("Fetching…")
        self._fetch_doi(doi)

    @work(thread=True)
    def _fetch_doi(self, doi: str) -> None:
        try:
            from bibtui.bib.doi import fetch_by_doi

            entry = fetch_by_doi(doi)
            self.app.call_from_thread(self._on_fetch_success, entry)
        except Exception as e:  # noqa: BLE001
            self.app.call_from_thread(self._on_fetch_error, str(e))

    def _on_fetch_success(self, entry: BibEntry) -> None:
        status = self.query_one("#doi-status", Static)
        status.set_classes("success")
        status.update(f"Found: {entry.title[:60]}")
        self.app.call_later(self._confirm, entry)

    def _on_fetch_error(self, message: str) -> None:
        status = self.query_one("#doi-status", Static)
        status.set_classes("error")
        status.update(f"Error: {message}")

    def _confirm(self, entry: BibEntry) -> None:
        self.dismiss(entry)

    def action_cancel(self) -> None:
        self.dismiss(None)


class EditModal(ModalScreen[BibEntry | None]):
    """Modal to edit key fields of an entry."""

    BINDINGS = [
        Binding("ctrl+s", "save_and_close", "Write", show=True),
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    EditModal {
        align: center middle;
    }
    EditModal > Vertical {
        width: 80;
        height: auto;
        max-height: 92%;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    EditModal #edit-fields {
        height: 22;
    }
    EditModal Input {
        margin-bottom: 1;
    }
    EditModal Label {
        color: $accent;
    }
    EditModal TextArea {
        height: 6;
        margin-bottom: 1;
    }
    """

    def __init__(self, entry: BibEntry, **kwargs):
        super().__init__(**kwargs)
        self._entry = entry

    def compose(self) -> ComposeResult:
        e = self._entry
        with Vertical():
            yield Label(
                f"[bold]Edit Entry[/bold]  [dim]{e.key}[/dim]", classes="modal-title"
            )
            with VerticalScroll(id="edit-fields"):
                yield Label("Title")
                yield Input(value=e.title, id="edit-title")
                yield Label("Author")
                yield Input(value=e.author, id="edit-author")
                yield Label("Year")
                yield Input(value=e.year, id="edit-year")
                yield Label("Journal / Booktitle")
                yield Input(value=e.journal, id="edit-journal")
                yield Label("DOI")
                yield Input(value=e.doi, id="edit-doi")
                yield Label("Keywords")
                yield Input(value=e.keywords, id="edit-keywords")
                yield Label("PDF file path")
                yield Input(value=e.file, id="edit-file")
                yield Label("Abstract")
                yield TextArea(e.abstract, id="edit-abstract")
                yield Label("Comment")
                yield TextArea(e.comment, id="edit-comment")
            with Horizontal(classes="modal-buttons"):
                yield Button("Write", variant="primary", id="btn-save")
                yield Button("Cancel", id="btn-cancel")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-save":
            self._save()

    def on_input_submitted(self, _: Input.Submitted) -> None:
        self._save()

    def _save(self) -> None:
        e = self._entry

        def v(widget_id: str) -> str:
            return self.query_one(widget_id, Input).value

        e.title = v("#edit-title")
        e.author = v("#edit-author")
        e.year = v("#edit-year")
        e.journal = v("#edit-journal")
        e.doi = v("#edit-doi")
        e.keywords = v("#edit-keywords")
        e.file = v("#edit-file")
        e.abstract = self.query_one("#edit-abstract", TextArea).text
        e.comment = self.query_one("#edit-comment", TextArea).text
        self.dismiss(e)

    def action_save_and_close(self) -> None:
        self._save()

    def action_cancel(self) -> None:
        self.dismiss(None)


class KeywordsModal(ModalScreen["tuple[str, set[str]] | None"]):
    """Keyword picker: select from all bib-wide keywords, add new ones."""

    BINDINGS = [
        Binding("ctrl+s", "save", "Write", show=True),
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    KeywordsModal {
        align: center middle;
    }
    KeywordsModal > Vertical {
        width: 70;
        height: 80%;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    KeywordsModal #kw-filter {
        margin-bottom: 1;
    }
    KeywordsModal SelectionList {
        height: 1fr;
        border: solid $panel;
    }
    KeywordsModal #kw-hints {
        height: auto;
        margin-top: 1;
        color: $text-muted;
    }
    """

    def __init__(
        self,
        entry: BibEntry,
        all_keywords: list[str],
        keyword_counts: dict[str, int],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._all_keywords = list(all_keywords)
        self._selected: set[str] = set(entry.keywords_list)
        self._shown: list[str] = []
        self._keyword_counts = keyword_counts
        self._delete_everywhere: set[str] = set()

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Edit Keywords[/bold]", classes="modal-title")
            yield Input(
                placeholder="Filter or type new keyword + Enter to add…", id="kw-filter"
            )
            yield SelectionList(id="kw-list")
            yield Static(
                "[dim]Esc close · Enter add new  |  ↓/↑ navigate · Space toggle · ⌫ delete everywhere[/dim]",
                id="kw-hints",
            )
            with Horizontal(classes="modal-buttons"):
                yield Button("Write", variant="primary", id="btn-save")
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self._rebuild_list("")
        self.call_after_refresh(self.query_one("#kw-filter", Input).focus)

    def on_key(self, event: events.Key) -> None:
        sl = self.query_one(SelectionList)
        kw_filter = self.query_one("#kw-filter", Input)
        if self.focused is kw_filter and event.key == "down":
            sl.focus()
            event.stop()
        elif self.focused is sl and event.key == "up" and sl.highlighted == 0:
            kw_filter.focus()
            event.stop()
        elif self.focused is sl and event.key == "backspace":
            self._delete_highlighted()
            event.stop()

    def _delete_highlighted(self) -> None:
        sl = self.query_one(SelectionList)
        highlighted = sl.highlighted
        if highlighted is None or highlighted >= len(self._shown):
            return
        kw = self._shown[highlighted]
        count = self._keyword_counts.get(kw, 0)
        noun = "entry" if count == 1 else "entries"
        msg = f"Remove '[bold]{kw}[/bold]' from all {count} {noun}?"
        self.app.push_screen(
            ConfirmModal(msg),
            lambda confirmed: self._on_delete_confirmed(confirmed, kw),
        )

    def _on_delete_confirmed(self, confirmed: bool | None, kw: str) -> None:
        if not confirmed:
            return
        self._delete_everywhere.add(kw)
        self._selected.discard(kw)
        if kw in self._all_keywords:
            self._all_keywords.remove(kw)
        filter_val = self.query_one("#kw-filter", Input).value
        self._rebuild_list(filter_val)

    def _sync_from_list(self) -> None:
        """Pull current checkbox state into self._selected."""
        sl = self.query_one(SelectionList)
        selected_now = set(sl.selected)
        for kw in self._shown:
            if kw in selected_now:
                self._selected.add(kw)
            else:
                self._selected.discard(kw)

    def _rebuild_list(self, filter_text: str) -> None:
        f = filter_text.lower()
        # Always show selected keywords first, then filtered rest
        selected_shown = sorted(kw for kw in self._selected if not f or f in kw.lower())
        rest = [
            kw
            for kw in self._all_keywords
            if kw not in self._selected and (not f or f in kw.lower())
        ]
        self._shown = selected_shown + rest
        sl = self.query_one(SelectionList)
        sl.clear_options()
        for kw in self._shown:
            sl.add_option(Selection(kw, kw, kw in self._selected))

    @on(Input.Changed, "#kw-filter")
    def on_filter_changed(self, event: Input.Changed) -> None:
        self._sync_from_list()
        self._rebuild_list(event.value)

    @on(Input.Submitted, "#kw-filter")
    def on_filter_submitted(self, event: Input.Submitted) -> None:
        kw = event.value.strip()
        if not kw:
            self.query_one(SelectionList).focus()
            return
        self._sync_from_list()
        if kw not in self._all_keywords:
            self._all_keywords.insert(0, kw)
        self._selected.add(kw)
        self.query_one("#kw-filter", Input).clear()
        self._rebuild_list("")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-save":
            self._save()

    def _save(self) -> None:
        self._sync_from_list()
        # Preserve original order from all_keywords, then any extras
        ordered = [kw for kw in self._all_keywords if kw in self._selected]
        self.dismiss((", ".join(ordered), self._delete_everywhere))

    def action_save(self) -> None:
        self._save()

    def action_cancel(self) -> None:
        self.dismiss(None)


class SettingsModal(ModalScreen["Config | None"]):
    """Settings dialog — currently just the PDF base directory."""

    BINDINGS = [
        Binding("ctrl+s", "save", "Write", show=True),
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    SettingsModal {
        align: center middle;
    }
    SettingsModal > Vertical {
        width: 70;
        height: auto;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    SettingsModal Input {
        margin-bottom: 1;
    }
    SettingsModal .setting-row {
        height: auto;
        align: left middle;
        margin-bottom: 1;
    }
    SettingsModal .setting-row Label {
        width: 1fr;
    }
    """

    def __init__(self, config: Config, **kwargs):
        super().__init__(**kwargs)
        self._config = config

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Settings[/bold]", classes="modal-title")
            yield Label("PDF base directory")
            yield Static(
                "[dim]Filenames in the file field are resolved relative to this path.[/dim]"
            )
            yield Input(
                value=self._config.pdf_base_dir,
                placeholder="/home/user/Papers",
                id="pdf-base-dir",
            )
            yield Static("")

            yield Label("Unpaywall email")
            yield Static(
                "[dim]Used for open-access PDF lookup via Unpaywall — no registration needed.[/dim]"
            )
            yield Input(
                value=self._config.unpaywall_email,
                placeholder="me@example.com",
                id="unpaywall-email",
            )
            yield Static("")

            yield Label("PDF download directory")
            yield Static(
                "[dim]PDFs listed when you press [bold]a[/bold] to add an existing PDF. Defaults to ~/Downloads.[/dim]"
            )
            yield Input(
                value=self._config.pdf_download_dir,
                placeholder=str(__import__("pathlib").Path.home() / "Downloads"),
                id="pdf-download-dir",
            )
            yield Static("")

            yield Label("Auto-fetch PDF on import")
            yield Static(
                "[dim]Automatically fetch the PDF after importing an entry by DOI or paste.[/dim]"
            )
            with Horizontal(classes="setting-row"):
                yield Switch(value=self._config.auto_fetch_pdf, id="auto-fetch-pdf")
            yield Static("")

            yield Label("Check for updates on startup")
            yield Static(
                "[dim]Checks PyPI at most once per day in the background and notifies when a newer stable release is available.[/dim]"
            )
            with Horizontal(classes="setting-row"):
                yield Switch(
                    value=self._config.check_for_updates, id="check-for-updates"
                )

            with Horizontal(classes="modal-buttons"):
                yield Button("Write", variant="primary", id="btn-save")
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self.call_after_refresh(self.query_one("#pdf-base-dir", Input).focus)

    def _collect(self) -> None:
        self._config.pdf_base_dir = self.query_one("#pdf-base-dir", Input).value.strip()
        self._config.unpaywall_email = self.query_one(
            "#unpaywall-email", Input
        ).value.strip()
        self._config.pdf_download_dir = self.query_one(
            "#pdf-download-dir", Input
        ).value.strip()
        self._config.auto_fetch_pdf = self.query_one("#auto-fetch-pdf", Switch).value
        self._config.check_for_updates = self.query_one(
            "#check-for-updates", Switch
        ).value

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-save":
            self._collect()
            self.dismiss(self._config)

    def on_input_submitted(self, _: Input.Submitted) -> None:
        self._collect()
        self.dismiss(self._config)

    def action_save(self) -> None:
        self._collect()
        self.dismiss(self._config)

    def action_cancel(self) -> None:
        self.dismiss(None)


class HelpModal(ModalScreen[None]):
    """Keybinding reference overlay."""

    BINDINGS = [
        Binding("escape", "dismiss_help", "Close", show=False),
        Binding("?", "dismiss_help", "Close", show=False),
    ]
    DEFAULT_CSS = """
    HelpModal { align: center middle; }
    HelpModal > Vertical {
        width: 74; height: 80%;
        border: double $accent; background: $surface; padding: 1 2;
    }
    HelpModal VerticalScroll { height: 1fr; }
    HelpModal #help-about { margin-bottom: 1; color: $text-muted; }
    """

    def _make_about(self) -> str:
        try:
            from bibtui import __version__

            version = __version__
        except Exception:
            version = "unknown"
        return (
            f"[bold]bibtui[/bold] v{version}  —  BibTeX TUI\n"
            "[dim]Author:[/dim] Thomas Gölles\n"
            "[dim]Web:[/dim] https://thomasgoelles.com\n"
            "[dim]Repo:[/dim]   https://github.com/tgoelles/bib_tui"
        )

    _KEYS = """\
[bold]── Core ──────────────────────────────[/bold]
  [bold]q[/bold]         Quit
  [bold]w[/bold]         Write
  [bold]s[/bold]         Search
  [bold]e[/bold]         Edit entry (field form or raw BibTeX)
  [bold]k[/bold]         Edit keywords
    [bold]m[/bold]         Maximize/restore table pane
    [dim]Press m again to restore split view.[/dim]
  [bold]v[/bold]         Toggle raw / formatted view

[bold]── Add new entry ─────────────────────[/bold]
  [bold]d[/bold]         Import entry by DOI (fetches metadata online)
  [bold]ctrl+v[/bold]    Paste a raw BibTeX entry from clipboard
  [dim]Both methods reject duplicate cite keys.[/dim]

[bold]── Delete entry ──────────────────────[/bold]
  [bold]Del / ⌫[/bold]   Delete the selected entry (confirmation required)

[bold]── Keywords modal ────────────────────[/bold]
  [bold]Enter[/bold]     Add typed keyword
  [bold]Space[/bold]     Toggle selected keyword on/off
  [bold]⌫[/bold]         Delete highlighted keyword from all entries
  [bold]↓ / ↑[/bold]     Move between filter and list

[bold]── Entry state ───────────────────────[/bold]
  [bold]r[/bold]         Cycle read state
  [bold]p[/bold]         Cycle priority
  [bold]␣[/bold]         Show PDF
  [bold]b[/bold]         Open URL in browser (validates http/https)
    [bold]Shift+b[/bold]   Search OpenAlex (title first, then DOI)
  [bold]f[/bold]         Fetch PDF and link it to the entry
  [bold]a[/bold]         Add an existing PDF to the library and link it

[bold]── Fetch PDF ─────────────────────────[/bold]
  Sources tried in order:
  [bold]1.[/bold] arXiv      — arXiv DOI or arxiv.org URL
  [bold]2.[/bold] Unpaywall  — OA by DOI (set email in Ctrl+P)
  [bold]3.[/bold] Direct URL — entry URL pointing to a PDF
  PDF saved to base directory from Settings.
  [dim]Some publishers block automated downloads.[/dim]

[bold]── Library actions ───────────────────[/bold]
    [bold]ctrl+p[/bold]    Open command palette
    [bold]Library: Fetch missing PDFs[/bold]
    [dim]Shows a toggle for: Overwrite broken links.[/dim]
    [bold]Library: Unify citekeys (AuthorYear)[/bold]
    [dim]Entries already matching AuthorYear are left unchanged.[/dim]
    [dim]Changing citekeys may break existing LaTeX documents.[/dim]

[bold]── Rating ────────────────────────────[/bold]
  [bold]1 – 5[/bold]     Set star rating
  [bold]0[/bold]         Mark unrated

[bold]── Other ─────────────────────────────[/bold]
    [bold]ctrl+c[/bold]    Copy selected text (or cite key if none focused)
  [bold]?[/bold]         Show this help
    [bold]ctrl+p[/bold]    Command palette (Settings + Library actions)
        [bold]maximize[/bold]  (palette) maximize focused pane
  [bold]Esc[/bold]       Clear search / close modal
  [dim]  Clipboard uses OSC 52 — requires a modern terminal[/dim]
  [dim]  In all modals: Ctrl+S = Write/Save, Esc = Cancel[/dim]

[bold]── Sorting ───────────────────────────[/bold]
  Click any column header to sort by that column.
  Click the same header again to reverse the order.
  Active sort column is marked with [bold]▲[/bold] (asc) or [bold]▼[/bold] (desc).
    Cols: [bold]◉[/bold] state  [bold]![/bold] prio  [bold]◫[/bold] PDF  [bold]🔗[/bold] URL  Type  Year  Author  Journal  Title  Added  [bold]★[/bold]
"""
    _SEARCH = """\
[bold]── Plain text ────────────────────────[/bold]
  Searches title, author, keywords, and key.
  Multiple tokens are ANDed (AND keyword optional).

[bold]── Field prefixes ────────────────────[/bold]
  [bold]a:[/bold] / [bold]author:[/bold]    filter by author
  [bold]t:[/bold] / [bold]title:[/bold]     filter by title
  [bold]j:[/bold] / [bold]journal:[/bold]   filter by journal
  [bold]k:[/bold] / [bold]kw:[/bold]        filter by keyword
  [bold]y:[/bold] / [bold]year:[/bold]      filter by year or range
  [bold]u:[/bold] / [bold]url:[/bold]       filter by URL
  [bold]c:[/bold] / [bold]citekey:[/bold]   filter by cite key

[bold]── Examples ──────────────────────────[/bold]
  [dim]glacier[/dim]                    all fields
  [dim]a:smith t:glacier[/dim]          combined
  [dim]j:nature AND y:2025[/dim]        journal + year
  [dim]y:2015-2023[/dim]                year range
  [dim]k:ice a:jones[/dim]              keyword + author
  [dim]c:smith2020[/dim]                exact cite key search"""

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Help[/bold]", classes="modal-title")
            with VerticalScroll():
                yield Static(self._make_about(), id="help-about")
                yield Label("[bold]Keybindings[/bold]", classes="modal-title")
                yield Static(self._KEYS)
                yield Label("[bold]Search syntax[/bold]", classes="modal-title")
                yield Static(self._SEARCH)
            with Horizontal(classes="modal-buttons"):
                yield Button("Close", variant="primary", id="btn-close")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss()

    def action_dismiss_help(self) -> None:
        self.dismiss()


class RawEditModal(ModalScreen[BibEntry | None]):
    """Edit a BibTeX entry as raw text."""

    BINDINGS = [
        Binding("ctrl+s", "save", "Write", show=True),
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    RawEditModal {
        align: center middle;
    }
    RawEditModal > Vertical {
        width: 90;
        height: 40;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    RawEditModal TextArea {
        height: 1fr;
    }
    #raw-edit-error {
        color: $error;
    }
    """

    def __init__(self, entry: BibEntry, **kwargs):
        super().__init__(**kwargs)
        self._entry = entry

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label(
                f"[bold]Edit Raw BibTeX[/bold]  [dim]{self._entry.key}[/dim]",
                classes="modal-title",
            )
            yield TextArea(
                entry_to_bibtex_str(self._entry),
                id="raw-edit-area",
            )
            yield Static("", id="raw-edit-error")
            with Horizontal(classes="modal-buttons"):
                yield Button("Write", variant="primary", id="btn-save")
                yield Button("Cancel", id="btn-cancel")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-save":
            self._save()

    def _save(self) -> None:
        text = self.query_one("#raw-edit-area", TextArea).text
        error = self.query_one("#raw-edit-error", Static)
        try:
            entry = bibtex_str_to_entry(text)
            self.dismiss(entry)
        except Exception as e:
            error.update(f"Parse error: {e}")

    def action_save(self) -> None:
        self._save()

    def action_cancel(self) -> None:
        self.dismiss(None)


class PasteModal(ModalScreen["BibEntry | None"]):
    """Modal to import a BibTeX entry from pasted clipboard text."""

    BINDINGS = [
        Binding("ctrl+s", "do_import", "Import", show=True),
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    PasteModal {
        align: center middle;
    }
    PasteModal > Vertical {
        width: 90;
        height: 40;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    PasteModal TextArea {
        height: 1fr;
    }
    #paste-error {
        color: $error;
    }
    """

    def __init__(self, text: str = "", **kwargs):
        super().__init__(**kwargs)
        self._text = text

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Paste BibTeX Entry[/bold]", classes="modal-title")
            yield TextArea(
                self._text,
                id="paste-area",
            )
            yield Static("", id="paste-error")
            with Horizontal(classes="modal-buttons"):
                yield Button("Import", variant="primary", id="btn-save")
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self.call_after_refresh(self.query_one("#paste-area", TextArea).focus)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-save":
            self._do_import()

    def _do_import(self) -> None:
        text = self.query_one("#paste-area", TextArea).text
        error = self.query_one("#paste-error", Static)
        error.update("")
        try:
            entry = bibtex_str_to_entry(text)
            self.dismiss(entry)
        except Exception as e:
            error.update(f"Parse error: {e}")

    def action_do_import(self) -> None:
        self._do_import()

    def action_cancel(self) -> None:
        self.dismiss(None)


class AddPDFModal(ModalScreen["str | None"]):
    """Pick an existing PDF from the download directory, filter by name, and link it."""

    BINDINGS = [
        Binding("ctrl+s", "add", "Add", show=True),
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    DEFAULT_CSS = """
    AddPDFModal {
        align: center middle;
    }
    AddPDFModal > Vertical {
        width: 80;
        height: 30;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    AddPDFModal Input {
        margin-bottom: 1;
    }
    AddPDFModal ListView {
        height: 1fr;
        border: solid $panel;
        margin-bottom: 1;
    }
    AddPDFModal #add-hint {
        color: $text-muted;
        margin-bottom: 1;
    }
    AddPDFModal #add-preview-hint {
        color: $text-muted;
        height: auto;
        margin-bottom: 1;
    }
    AddPDFModal #add-error {
        color: $error;
    }
    """

    def __init__(
        self,
        entry: BibEntry,
        base_dir: str,
        download_dir: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._entry = entry
        self._base_dir = base_dir
        from pathlib import Path

        self._download_dir = download_dir or str(Path.home() / "Downloads")
        self._all_pdfs: list = []
        self._filtered: list = []

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label(
                f"[bold]Add PDF[/bold]  [dim]{self._entry.key}[/dim]",
                classes="modal-title",
            )
            yield Static("", id="add-hint")
            yield Input(placeholder="type to filter…", id="add-filter")
            yield ListView(id="add-list")
            yield Static(
                "[dim]↓/↑ navigate · Space preview [/dim]",
                id="add-preview-hint",
            )
            yield Static("", id="add-error")
            with Horizontal(classes="modal-buttons"):
                yield Button("Add", variant="primary", id="btn-add")
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self._scan()
        self.call_after_refresh(self.query_one("#add-filter", Input).focus)

    def _scan(self) -> None:
        from pathlib import Path

        dl = Path(self._download_dir).expanduser()
        hint = self.query_one("#add-hint", Static)
        if not dl.is_dir():
            hint.update(
                f"[dim]Download dir not found: {dl}  ·  enter a path manually[/dim]"
            )
            self._all_pdfs = []
        else:
            pdfs = sorted(
                dl.glob("*.pdf"), key=lambda p: p.stat().st_mtime, reverse=True
            )
            self._all_pdfs = pdfs
            hint.update(
                f"[dim]{dl}  ·  {len(pdfs)} PDF{'s' if len(pdfs) != 1 else ''}[/dim]"
            )
        self._filtered = list(self._all_pdfs)
        self._refresh_list()

    def _refresh_list(self) -> None:
        lv = self.query_one(ListView)
        lv.clear()
        for p in self._filtered:
            stat = p.stat()
            size = stat.st_size
            size_str = (
                f"{size / 1048576:.1f} MB"
                if size >= 1048576
                else f"{size / 1024:.0f} KB"
            )
            age = _format_age(stat.st_mtime)
            lv.append(ListItem(Label(f"{p.name}  [dim]{size_str}  {age}[/dim]")))

    @on(Input.Changed, "#add-filter")
    def _on_filter(self, event: Input.Changed) -> None:
        q = event.value.strip().lower()
        self._filtered = (
            [p for p in self._all_pdfs if q in p.name.lower()]
            if q
            else list(self._all_pdfs)
        )
        self._refresh_list()

    def on_key(self, event: events.Key) -> None:
        """Down in the Input moves focus to the list; Up from the first item returns focus."""
        lv = self.query_one(ListView)
        inp = self.query_one("#add-filter", Input)
        if self.focused is inp and event.key == "down" and self._filtered:
            lv.focus()
            event.stop()
        elif self.focused is lv and event.key == "up" and (lv.index or 0) == 0:
            inp.focus()
            event.stop()
        elif self.focused is lv and event.key == "space":
            self._preview_selected()
            event.stop()

    def _preview_selected(self) -> None:
        import platform
        import subprocess

        lv = self.query_one(ListView)
        idx = lv.index
        if idx is None or idx >= len(self._filtered):
            return
        path = self._filtered[idx]
        try:
            if platform.system() == "Darwin":
                subprocess.Popen(["open", str(path)])
            else:
                subprocess.Popen(["xdg-open", str(path)])
        except Exception as e:
            self.query_one("#add-error", Static).update(f"Could not open: {e}")

    @on(Input.Submitted, "#add-filter")
    def _on_filter_submitted(self, _: Input.Submitted) -> None:
        self._confirm()

    @on(ListView.Selected)
    def _on_list_selected(self, event: ListView.Selected) -> None:
        idx = self.query_one(ListView).index
        if idx is not None and idx < len(self._filtered):
            self._add_path(self._filtered[idx])

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-add":
            self._confirm()

    def _confirm(self) -> None:
        from pathlib import Path

        lv = self.query_one(ListView)
        idx = lv.index
        if self._filtered and idx is not None and idx < len(self._filtered):
            self._add_path(self._filtered[idx])
        else:
            # Fallback: treat the filter text as a custom path
            val = self.query_one("#add-filter", Input).value.strip()
            if val:
                self._add_path(Path(val))
            else:
                self.query_one("#add-error", Static).update(
                    "Select a file or enter a path."
                )

    def _add_path(self, src) -> None:
        from pathlib import Path

        from bibtui.pdf.fetcher import FetchError, add_pdf

        error = self.query_one("#add-error", Static)
        error.update("")
        try:
            dest = add_pdf(Path(src), self._entry, self._base_dir)
            self.dismiss(str(dest))
        except FetchError as exc:
            error.update(str(exc))

    def action_add(self) -> None:
        self._confirm()

    def action_cancel(self) -> None:
        self.dismiss(None)


class FetchPDFModal(ModalScreen["str | None"]):
    """Fetch a PDF for an entry in a background thread and show progress."""

    BINDINGS = [Binding("escape", "cancel", "Cancel", show=False)]

    DEFAULT_CSS = """
    FetchPDFModal {
        align: center middle;
    }
    FetchPDFModal > Vertical {
        width: 70;
        height: auto;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    FetchPDFModal LoadingIndicator {
        height: 3;
    }
    FetchPDFModal #fetch-status {
        margin-top: 1;
        margin-bottom: 1;
        color: $text;
    }
    FetchPDFModal #fetch-status.success {
        color: $success;
    }
    FetchPDFModal #fetch-status.error {
        color: $error;
    }
    """

    def __init__(
        self,
        entry: BibEntry,
        dest_dir: str,
        unpaywall_email: str = "",
        overwrite: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._entry = entry
        self._dest_dir = dest_dir
        self._email = unpaywall_email
        self._overwrite = overwrite
        self._saved_path: str | None = None

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label(
                f"[bold]Fetch PDF[/bold]  [dim]{self._entry.key}[/dim]",
                classes="modal-title",
            )
            yield LoadingIndicator(id="fetch-loading")
            yield Static("", id="fetch-status")
            with Horizontal(classes="modal-buttons"):
                yield Button("Close", variant="primary", id="btn-close", disabled=True)
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self._do_fetch()

    @work(thread=True)
    def _do_fetch(self) -> None:
        from bibtui.pdf.fetcher import FetchError, fetch_pdf

        try:
            path = fetch_pdf(
                self._entry, self._dest_dir, self._email, overwrite=self._overwrite
            )
            self.app.call_from_thread(self._on_success, path)
        except FetchError as exc:
            self.app.call_from_thread(self._on_error, str(exc))
        except Exception as exc:  # noqa: BLE001
            self.app.call_from_thread(self._on_error, f"Unexpected error: {exc}")

    def _on_success(self, path: str) -> None:
        self.query_one("#fetch-loading", LoadingIndicator).display = False
        status = self.query_one("#fetch-status", Static)
        status.set_class(True, "success")
        status.set_class(False, "error")
        status.update(f"Saved PDF:\n{path}")
        self.query_one("#btn-close", Button).disabled = False
        self.query_one("#btn-cancel", Button).disabled = True
        self._saved_path = path

    def _on_error(self, message: str) -> None:
        self.query_one("#fetch-loading", LoadingIndicator).display = False
        status = self.query_one("#fetch-status", Static)
        status.set_class(False, "success")
        status.set_class(True, "error")
        status.update(self._format_fetch_error(message))
        self.query_one("#btn-close", Button).disabled = False

    def _format_fetch_error(self, message: str) -> str:
        title = "Could not fetch PDF for this entry."
        lines = [line.strip() for line in message.splitlines() if line.strip()]
        if not lines:
            return title

        reasons = [
            line[1:].strip() if line.startswith("•") else line
            for line in lines
            if line.startswith("•") or line.startswith("-")
        ]
        if not reasons:
            reasons = [
                line
                for line in lines
                if not line.lower().startswith("could not fetch pdf")
            ]

        if not reasons:
            return title

        formatted_reasons = "\n".join(f"• {reason}" for reason in reasons)
        return f"{title}\n\n{formatted_reasons}"

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.dismiss(None)
        elif event.button.id == "btn-close":
            self.dismiss(self._saved_path)

    def action_cancel(self) -> None:
        self.dismiss(None)


class BatchFetchPDFModal(ModalScreen["dict | None"]):
    """Fetch PDFs for many entries in a background thread and show progress."""

    BINDINGS = [Binding("escape", "cancel", "Cancel", show=False)]

    DEFAULT_CSS = """
    BatchFetchPDFModal {
        align: center middle;
    }
    BatchFetchPDFModal > Vertical {
        width: 72;
        height: auto;
        border: double $accent;
        background: $surface;
        padding: 1 2;
    }
    BatchFetchPDFModal LoadingIndicator {
        height: 3;
    }
    BatchFetchPDFModal #batch-fetch-progress,
    BatchFetchPDFModal #batch-fetch-status {
        margin-top: 1;
        color: $text;
    }
    BatchFetchPDFModal #batch-fetch-status.success {
        color: $success;
    }
    BatchFetchPDFModal #batch-fetch-status.error {
        color: $error;
    }
    """

    def __init__(
        self,
        entries: list[BibEntry],
        dest_dir: str,
        unpaywall_email: str = "",
        overwrite_broken_links: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._entries = entries
        self._dest_dir = dest_dir
        self._email = unpaywall_email
        self._overwrite_broken_links = overwrite_broken_links
        self._cancel_requested = False
        self._done = False
        self._result: dict | None = None

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("[bold]Fetch Missing PDFs[/bold]", classes="modal-title")
            yield LoadingIndicator(id="batch-fetch-loading")
            yield Static("Preparing…", id="batch-fetch-progress")
            yield Static("", id="batch-fetch-status")
            with Horizontal(classes="modal-buttons"):
                yield Button("Close", variant="primary", id="btn-close", disabled=True)
                yield Button("Cancel", id="btn-cancel")

    def on_mount(self) -> None:
        self._do_fetch()

    @work(thread=True)
    def _do_fetch(self) -> None:
        from bibtui.pdf.fetcher import FetchError, fetch_pdf

        total = len(self._entries)
        success = 0
        failed = 0
        skipped = 0
        canceled = False
        paths_by_key: dict[str, str] = {}
        failures: list[str] = []

        for index, entry in enumerate(self._entries, start=1):
            if self._cancel_requested:
                canceled = True
                break

            self.app.call_from_thread(
                self._on_progress,
                f"[{index}/{total}] {entry.key}",
            )

            if not entry.doi and not entry.url:
                skipped += 1
                failures.append(f"{entry.key}: no DOI or URL")
                continue

            try:
                path = fetch_pdf(
                    entry,
                    self._dest_dir,
                    self._email,
                    overwrite=False,
                )
                paths_by_key[entry.key] = path
                success += 1
            except FetchError as exc:
                failed += 1
                failures.append(f"{entry.key}: {exc}")
            except Exception as exc:  # noqa: BLE001
                failed += 1
                failures.append(f"{entry.key}: unexpected error: {exc}")

        processed = success + failed + skipped
        self.app.call_from_thread(
            self._on_done,
            {
                "paths_by_key": paths_by_key,
                "processed": processed,
                "total": total,
                "success": success,
                "failed": failed,
                "skipped": skipped,
                "canceled": canceled,
                "failures": failures,
            },
        )

    def _on_progress(self, message: str) -> None:
        self.query_one("#batch-fetch-progress", Static).update(message)

    def _on_done(self, result: dict) -> None:
        self._done = True
        self._result = result
        self.query_one("#batch-fetch-loading", LoadingIndicator).display = False
        status = self.query_one("#batch-fetch-status", Static)
        if result["success"] > 0:
            status.set_class(True, "success")
            status.set_class(False, "error")
        else:
            status.set_class(False, "success")
            status.set_class(True, "error")

        canceled_text = " (canceled)" if result["canceled"] else ""
        status.update(
            "Finished"
            f"{canceled_text}: {result['success']} fetched, {result['failed']} failed, "
            f"{result['skipped']} skipped."
        )
        self.query_one("#batch-fetch-progress", Static).update(
            f"Processed {result['processed']} of {result['total']} entries."
        )
        self.query_one("#btn-close", Button).disabled = False
        self.query_one("#btn-cancel", Button).disabled = True

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "btn-cancel":
            self.action_cancel()
        elif event.button.id == "btn-close":
            self.dismiss(self._result)

    def action_cancel(self) -> None:
        if self._done:
            self.dismiss(self._result)
            return
        self._cancel_requested = True
        self.query_one("#batch-fetch-status", Static).update(
            "Stopping after current entry…"
        )
        self.query_one("#btn-cancel", Button).disabled = True


class FirstRunModal(ModalScreen[bool]):
    """One-time welcome notice — shown only on the very first launch."""

    BINDINGS = [Binding("escape", "got_it", "Got it", show=False)]

    DEFAULT_CSS = """
    FirstRunModal { align: center middle; }
    FirstRunModal > Vertical {
        width: 70; height: auto;
        border: double $accent; background: $surface; padding: 2 3;
    }
    FirstRunModal #welcome-title { text-align: center; margin-bottom: 1; }
    FirstRunModal #welcome-body  { margin-bottom: 1; }
    FirstRunModal .modal-buttons { height: 3; align: right middle; }
    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Static(
                "[bold]Welcome to bibtui![/bold]",
                id="welcome-title",
            )
            yield Static(
                "A terminal UI with keyboard and mouse support for your BibTeX libraries.\n\n"
                "[bold]You're ready to go[/bold] — no setup required.\n"
                "Just run  [bold]bibtui yourfile.bib[/bold]  and start browsing.\n\n"
                "[dim]Optional:[/dim] PDF fetching and attaching require a few"
                " extra settings\n"
                "(PDF dir, download dir, Unpaywall email).  Open them any time\n"
                "via [bold]Ctrl+P → Settings[/bold] or press [bold]Configure PDF features[/bold] below.",
                id="welcome-body",
            )
            with Horizontal(classes="modal-buttons"):
                yield Button("Got it", variant="primary", id="btn-got-it")
                yield Button(
                    "Configure PDF features", variant="default", id="btn-configure"
                )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(event.button.id == "btn-configure")

    def action_got_it(self) -> None:
        self.dismiss(False)



from datetime import datetime

DATE_ADDED_KEYS: tuple[str, ...] = ("date-added", "date_added", "dateadded")


def now_date_added_value() -> str:
    """Return canonical timestamp used for newly added entries."""
    return datetime.now().strftime("%Y-%m-%dT%H:%M:%S")


def extract_date_added(raw_fields: dict[str, str]) -> str:
    """Return the first known raw date-added field value, if present."""
    for key in DATE_ADDED_KEYS:
        value = (raw_fields.get(key) or "").strip()
        if value:
            return value
    return ""


def parse_bib_date(value: str) -> datetime | None:
    """Parse a bibliography date string into a datetime if possible."""
    text = value.strip()
    if not text:
        return None

    normalized = text.replace("Z", "+00:00")
    try:
        return datetime.fromisoformat(normalized)
    except ValueError:
        pass

    for fmt in ("%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y/%m/%d"):
        try:
            return datetime.strptime(text, fmt)
        except ValueError:
            continue
    return None


def format_bib_date(value: str, empty: str = "") -> str:
    """Format a bibliography date to compact YYYY-MM-DD for table display."""
    parsed = parse_bib_date(value)
    if parsed is not None:
        return parsed.strftime("%Y-%m-%d")
    return value[:10] if value else empty



import bibtexparser
from bibtexparser import model as bpmodel

from .models import READ_STATES, BibEntry


def _field_str(entry: bpmodel.Entry, key: str) -> str:
    """Extract string value from a bibtexparser Entry field, stripping outer braces.

    Field name lookup is case-insensitive: ``AUTHOR`` and ``author`` both work.
    """
    # bibtexparser v2 preserves original casing — search case-insensitively
    f = entry.fields_dict.get(key) or entry.fields_dict.get(key.upper())
    if f is None:
        # Full case-insensitive scan as final fallback
        key_lower = key.lower()
        for k, v in entry.fields_dict.items():
            if k.lower() == key_lower:
                f = v
                break
    if f is None:
        return ""
    val = f.value
    if isinstance(val, str):
        val = val.strip()
        # Strip outer curly braces that BibTeX uses for case protection
        if val.startswith("{") and val.endswith("}"):
            val = val[1:-1]
        return val
    return str(val).strip()


def _to_bib_entry(entry: bpmodel.Entry) -> BibEntry:
    known = {
        "title",
        "author",
        "year",
        "journal",
        "doi",
        "url",
        "abstract",
        "keywords",
        "comment",
        "ranking",
        "readstatus",
        "priority",
        "file",
    }
    raw = {}
    for k, f in entry.fields_dict.items():
        # Normalise to lowercase so raw_fields are always consistent
        k_norm = k.lower()
        if k_norm not in known:
            val = f.value
            raw[k_norm] = val if isinstance(val, str) else str(val)

    ranking_str = _field_str(entry, "ranking")  # JabRef format: rank1..rank5
    try:
        rating = (
            max(0, min(5, int(ranking_str.removeprefix("rank")))) if ranking_str else 0
        )
    except ValueError:
        rating = 0

    priority_str = _field_str(entry, "priority")  # JabRef format: prio1..prio3
    try:
        priority = (
            max(0, min(3, int(priority_str.removeprefix("prio"))))
            if priority_str
            else 0
        )
    except ValueError:
        priority = 0

    read_state = _field_str(entry, "readstatus")  # JabRef field name
    if read_state not in READ_STATES:
        read_state = ""

    return BibEntry(
        key=entry.key,
        entry_type=entry.entry_type,
        title=_field_str(entry, "title"),
        author=_field_str(entry, "author"),
        year=_field_str(entry, "year"),
        journal=_field_str(entry, "journal"),
        doi=_field_str(entry, "doi"),
        url=_field_str(entry, "url"),
        abstract=_field_str(entry, "abstract"),
        keywords=_field_str(entry, "keywords"),
        comment=_field_str(entry, "comment"),
        rating=rating,
        read_state=read_state,
        priority=priority,
        file=_field_str(entry, "file"),
        raw_fields=raw,
    )


def _to_bp_entry(entry: BibEntry) -> bpmodel.Entry:
    fields: list[bpmodel.Field] = []

    def add(key: str, value: str) -> None:
        if value:
            fields.append(bpmodel.Field(key=key, value=value))

    add("title", entry.title)
    add("author", entry.author)
    add("year", entry.year)
    add("journal", entry.journal)
    add("doi", entry.doi)
    add("url", entry.url)
    add("abstract", entry.abstract)
    add("keywords", entry.keywords)
    add("comment", entry.comment)

    if entry.rating:
        add("ranking", f"rank{entry.rating}")
    if entry.read_state:
        add("readstatus", entry.read_state)
    if entry.priority:
        add("priority", f"prio{entry.priority}")
    if entry.file:
        add("file", entry.file)

    for k, v in entry.raw_fields.items():
        if v:
            fields.append(bpmodel.Field(key=k, value=v))

    return bpmodel.Entry(key=entry.key, entry_type=entry.entry_type, fields=fields)


def entry_to_bibtex_str(entry: BibEntry) -> str:
    """Serialize a single BibEntry to a BibTeX string."""
    lib = bibtexparser.Library()
    lib.add(_to_bp_entry(entry))
    return bibtexparser.write_string(lib)


def bibtex_str_to_entry(text: str) -> BibEntry:
    """Parse a BibTeX string containing a single entry back to a BibEntry.

    Raises ValueError if parsing fails or no entry is found.
    """
    lib = bibtexparser.parse_string(text)
    if not lib.entries:
        raise ValueError("No valid BibTeX entry found in the text.")
    return _to_bib_entry(lib.entries[0])


def load(path: str) -> list[BibEntry]:
    lib = bibtexparser.parse_file(path)
    return [_to_bib_entry(e) for e in lib.entries]


def save(entries: list[BibEntry], path: str) -> None:
    lib = bibtexparser.Library()
    for entry in entries:
        lib.add(_to_bp_entry(entry))
    bibtexparser.write_file(path, lib)



import json
import urllib.request
from datetime import UTC, datetime, timedelta

from packaging.version import InvalidVersion, Version

PACKAGE_NAME = "bibtui"
PYPI_JSON_URL = f"https://pypi.org/pypi/{PACKAGE_NAME}/json"
CHECK_INTERVAL = timedelta(days=1)


def utc_now() -> datetime:
    return datetime.now(UTC)


def to_utc_iso(value: datetime) -> str:
    return (
        value.astimezone(UTC).replace(microsecond=0).isoformat().replace("+00:00", "Z")
    )


def parse_utc_iso(value: str) -> datetime | None:
    if not value:
        return None
    text = value.strip()
    if text.endswith("Z"):
        text = text[:-1] + "+00:00"
    try:
        parsed = datetime.fromisoformat(text)
    except ValueError:
        return None
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=UTC)
    return parsed.astimezone(UTC)


def is_due(last_check_utc: str, now: datetime | None = None) -> bool:
    now_utc = now or utc_now()
    last = parse_utc_iso(last_check_utc)
    if last is None:
        return True
    return (now_utc - last) >= CHECK_INTERVAL


def notified_today(last_notified_utc: str, now: datetime | None = None) -> bool:
    now_utc = now or utc_now()
    last = parse_utc_iso(last_notified_utc)
    if last is None:
        return False
    return last.date() == now_utc.date()


def is_newer_version(installed: str, latest: str) -> bool:
    try:
        return Version(latest) > Version(installed)
    except InvalidVersion:
        return False


def _stable_versions(releases: dict) -> list[Version]:
    stable: list[Version] = []
    for value in releases.keys():
        try:
            version = Version(value)
        except InvalidVersion:
            continue
        if version.is_prerelease or version.is_devrelease:
            continue
        stable.append(version)
    return stable


def fetch_latest_stable_version(timeout: int = 3) -> str | None:
    req = urllib.request.Request(
        PYPI_JSON_URL,
        headers={"User-Agent": "bibtui update-check"},
    )
    try:
        with urllib.request.urlopen(req, timeout=timeout) as response:
            data = json.loads(response.read().decode("utf-8"))
    except Exception:
        return None

    releases = data.get("releases")
    if not isinstance(releases, dict):
        return None

    stable = _stable_versions(releases)
    if not stable:
        return None
    return str(max(stable))



import tomllib
from dataclasses import dataclass
from pathlib import Path

CONFIG_PATH = Path.home() / ".config" / "bibtui" / "config.toml"


@dataclass
class Config:
    pdf_base_dir: str = ""
    unpaywall_email: str = ""
    pdf_download_dir: str = ""
    auto_fetch_pdf: bool = True
    update_last_check_utc: str = ""
    update_last_notified_utc: str = ""
    update_latest_version: str = ""
    check_for_updates: bool = True


def is_first_run() -> bool:
    """Return True if no config file exists yet (new installation)."""
    return not CONFIG_PATH.exists()


def _git_email() -> str:
    """Read user.email from global git config, or return empty string."""
    import subprocess

    try:
        result = subprocess.run(
            ["git", "config", "--global", "user.email"],
            capture_output=True,
            text=True,
            timeout=2,
        )
        return result.stdout.strip()
    except Exception:
        return ""


def default_config() -> Config:
    """Return a Config pre-filled with sensible platform defaults."""
    home = Path.home()
    return Config(
        pdf_base_dir=str(home / "Documents" / "papers"),
        unpaywall_email=_git_email(),
        pdf_download_dir=str(home / "Downloads"),
    )


def load_config() -> Config:
    if not CONFIG_PATH.exists():
        return default_config()
    try:
        with open(CONFIG_PATH, "rb") as f:
            data = tomllib.load(f)
    except (OSError, tomllib.TOMLDecodeError):
        return default_config()
    pdf = data.get("pdf", {})
    updates = data.get("updates", {})
    return Config(
        pdf_base_dir=pdf.get("base_dir", ""),
        unpaywall_email=pdf.get("unpaywall_email", ""),
        pdf_download_dir=pdf.get("download_dir", ""),
        auto_fetch_pdf=pdf.get("auto_fetch_pdf", True),
        update_last_check_utc=updates.get("last_check_utc", ""),
        update_last_notified_utc=updates.get("last_notified_utc", ""),
        update_latest_version=updates.get("latest_version", ""),
        check_for_updates=updates.get("check_for_updates", True),
    )


def _toml_escape(value: str) -> str:
    """Escape a string value for embedding in a TOML double-quoted string."""
    return value.replace("\\", "\\\\").replace('"', '\\"')


def save_config(config: Config) -> None:
    CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    lines = [
        "[pdf]",
        f'base_dir = "{_toml_escape(config.pdf_base_dir)}"',
        f'unpaywall_email = "{_toml_escape(config.unpaywall_email)}"',
        f'download_dir = "{_toml_escape(config.pdf_download_dir)}"',
        f"auto_fetch_pdf = {'true' if config.auto_fetch_pdf else 'false'}",
        "",
        "[updates]",
        f'last_check_utc = "{_toml_escape(config.update_last_check_utc)}"',
        f'last_notified_utc = "{_toml_escape(config.update_last_notified_utc)}"',
        f'latest_version = "{_toml_escape(config.update_latest_version)}"',
        f"check_for_updates = {'true' if config.check_for_updates else 'false'}",
        "",
    ]
    CONFIG_PATH.write_text("\n".join(lines), encoding="utf-8")



from bibtui.pdf.fetcher import FetchError, add_pdf, fetch_pdf, pdf_filename
from bibtui.pdf.paths import find_pdf_for_entry, format_jabref_path, parse_jabref_path

__all__ = [
    "FetchError",
    "add_pdf",
    "fetch_pdf",
    "pdf_filename",
    "find_pdf_for_entry",
    "format_jabref_path",
    "parse_jabref_path",
]



"""Detect the OS / Omarchy theme and map it to a Textual theme name."""

import os
from pathlib import Path

# Omarchy theme name  →  Textual built-in theme name
_THEME_MAP: dict[str, str] = {
    "catppuccin": "catppuccin-mocha",
    "catppuccin-latte": "catppuccin-latte",
    "ethereal": "textual-dark",
    "everforest": "gruvbox",
    "flexoki-light": "flexoki",
    "gruvbox": "gruvbox",
    "hackerman": "monokai",
    "kanagawa": "tokyo-night",
    "matte-black": "textual-dark",
    "nord": "nord",
    "osaka-jade": "tokyo-night",
    "ristretto": "rose-pine",
    "rose-pine": "rose-pine",
    "tokyo-night": "tokyo-night",
}

_OMARCHY_THEME_LINK = Path.home() / ".config" / "omarchy" / "current" / "theme"


def _omarchy_theme_name() -> str | None:
    """Return the active Omarchy theme name by reading the symlink, or None."""
    try:
        target = os.readlink(_OMARCHY_THEME_LINK)
        return Path(target).name
    except OSError:
        return None


def _omarchy_is_light() -> bool:
    """True if the current Omarchy theme ships a light.mode marker file."""
    return (_OMARCHY_THEME_LINK / "light.mode").exists()


def detect_theme() -> str:
    """Return a Textual theme name that matches the current OS/Omarchy theme.

    Falls back to ``textual-dark`` when Omarchy is not installed or the theme
    has no direct mapping.
    """
    name = _omarchy_theme_name()
    if name is None:
        return "textual-dark"
    if name in _THEME_MAP:
        return _THEME_MAP[name]
    # Unknown Omarchy theme — at least honour light vs dark.
    return "textual-light" if _omarchy_is_light() else "textual-dark"



from habanero import Crossref

from bibtui.utils.dates import now_date_added_value

from .citekeys import author_year_base
from .models import BibEntry


def fetch_by_doi(doi: str) -> BibEntry:
    cr = Crossref()
    result = cr.works(ids=doi)
    msg = result["message"]

    def get_str(key: str) -> str:
        val = msg.get(key, "")
        if isinstance(val, list):
            return val[0] if val else ""
        return str(val) if val else ""

    # Authors
    authors = msg.get("author", [])
    author_parts = []
    for a in authors:
        family = a.get("family", "")
        given = a.get("given", "")
        if family and given:
            author_parts.append(f"{family}, {given}")
        elif family:
            author_parts.append(family)
    author_str = " and ".join(author_parts)

    # Year
    year = ""
    date = (
        msg.get("published-print") or msg.get("published-online") or msg.get("issued")
    )
    if date:
        parts = date.get("date-parts", [[]])
        if parts and parts[0]:
            year = str(parts[0][0])

    # Title
    titles = msg.get("title", [])
    title = titles[0] if titles else ""

    # Journal
    container = msg.get("container-title", [])
    journal = container[0] if container else ""

    # Entry type
    crossref_type = msg.get("type", "journal-article")
    type_map = {
        "journal-article": "article",
        "proceedings-article": "inproceedings",
        "book": "book",
        "book-chapter": "incollection",
        "dissertation": "phdthesis",
        "report": "techreport",
        "dataset": "misc",
        "posted-content": "misc",
    }
    entry_type = type_map.get(crossref_type, "misc")

    # Build citation key: AuthorYear (normalized)
    key = author_year_base(author_str, year)

    raw: dict[str, str] = {
        "date-added": now_date_added_value(),
    }
    if msg.get("volume"):
        raw["volume"] = str(msg["volume"])
    if msg.get("issue"):
        raw["number"] = str(msg["issue"])
    page = msg.get("page", "")
    if page:
        raw["pages"] = str(page)
    publisher = msg.get("publisher", "")
    if publisher:
        raw["publisher"] = publisher

    return BibEntry(
        key=key,
        entry_type=entry_type,
        title=title,
        author=author_str,
        year=year,
        journal=journal,
        doi=doi,
        raw_fields=raw,
    )



import os
import platform
import re
import subprocess
import webbrowser
from typing import cast
from urllib.parse import quote_plus, urlparse

from textual import events, on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.command import DiscoveryHit, Hit, Hits, Provider
from textual.containers import Horizontal
from textual.widgets import DataTable, Footer, Header, Input, TextArea

from bibtui import __version__
from bibtui.bib import parser
from bibtui.bib.citekeys import (
    author_year_base,
    is_canonical_author_year_key,
    make_unique_key,
)
from bibtui.bib.models import BibEntry
from bibtui.pdf.fetcher import pdf_filename
from bibtui.pdf.paths import find_pdf_for_entry, format_jabref_path, parse_jabref_path
from bibtui.utils import update_check
from bibtui.utils.config import (
    CONFIG_PATH,
    Config,
    is_first_run,
    load_config,
    save_config,
)
from bibtui.utils.theme import detect_theme
from bibtui.widgets.entry_detail import EntryDetail
from bibtui.widgets.entry_list import EntryList
from bibtui.widgets.modals import (
    AddPDFModal,
    BatchFetchPDFModal,
    ConfirmModal,
    DOIModal,
    EditModal,
    FetchPDFModal,
    FirstRunModal,
    HelpModal,
    KeywordsModal,
    LibraryFetchConfirmModal,
    PasteModal,
    RawEditModal,
    SettingsModal,
)


class SettingsProvider(Provider):
    """Exposes the Settings dialog through the command palette."""

    async def discover(self) -> Hits:
        app = cast("BibTuiApp", self.app)
        yield DiscoveryHit(
            "Settings", app.action_settings, help="Open the settings dialog"
        )

    async def search(self, query: str) -> Hits:
        app = cast("BibTuiApp", self.app)
        matcher = self.matcher(query)
        score = matcher.match("Settings")
        if score > 0:
            yield Hit(
                score,
                matcher.highlight("Settings"),
                app.action_settings,
                help="Open the settings dialog",
            )


class LibraryProvider(Provider):
    """Exposes library-wide actions through the command palette."""

    async def discover(self) -> Hits:
        app = cast("BibTuiApp", self.app)
        yield DiscoveryHit(
            "Library: Fetch missing PDFs",
            app.action_fetch_missing_pdfs,
            help="Fetch PDFs for entries missing local files",
        )
        yield DiscoveryHit(
            "Library: Unify citekeys (AuthorYear)",
            app.action_unify_citekeys,
            help="Unify citekeys to AuthorYear format",
        )

    async def search(self, query: str) -> Hits:
        app = cast("BibTuiApp", self.app)
        matcher = self.matcher(query)
        for label, action, help_text in (
            (
                "Library: Fetch missing PDFs",
                app.action_fetch_missing_pdfs,
                "Fetch PDFs for entries missing local files",
            ),
            (
                "Library: Unify citekeys (AuthorYear)",
                app.action_unify_citekeys,
                "Unify citekeys to AuthorYear format",
            ),
        ):
            score = matcher.match(label)
            if score > 0:
                yield Hit(score, matcher.highlight(label), action, help=help_text)


class BibTuiApp(App):
    """BibTeX TUI Application."""

    COMMANDS = App.COMMANDS | {SettingsProvider, LibraryProvider}

    CSS_PATH = "bibtui.tcss"

    BINDINGS = [
        # Core
        Binding("q", "quit", "Quit"),
        Binding("w", "save", "Write"),
        Binding("s", "focus_search", "Search"),
        Binding("e", "edit_entry", "Edit"),
        Binding("d", "doi_import", "From DOI"),
        Binding("k", "edit_keywords", "Keywords"),
        Binding("m", "toggle_table_maximize", "Max table"),
        Binding("v", "toggle_view", "View"),
        # Entry state
        Binding("r", "cycle_read_state", "State"),
        Binding("p", "cycle_priority", "Prio"),
        Binding("space", "open_pdf", "␣ Show PDF"),
        Binding("b", "open_url", "Browser"),
        Binding("B", "open_openalex", "OpenAlex"),
        Binding("f", "fetch_pdf", "Fetch PDF"),
        Binding("a", "add_pdf", "Add PDF"),
        # Rating (hidden from footer)
        Binding("0", "set_rating('0')", "Unrated", show=False),
        Binding("1", "set_rating('1')", "★", show=False),
        Binding("2", "set_rating('2')", "★★", show=False),
        Binding("3", "set_rating('3')", "★★★", show=False),
        Binding("4", "set_rating('4')", "★★★★", show=False),
        Binding("5", "set_rating('5')", "★★★★★", show=False),
        # Copy
        Binding("ctrl+c", "copy_key", "Copy key", show=False, priority=True),
        # Delete
        Binding("delete", "delete_entry", "Delete", show=False),
        Binding("backspace", "delete_entry", "Delete", show=False),
        # Help
        Binding("?", "show_help", "Help"),
        Binding("escape", "clear_search", "Clear search", show=False),
    ]

    def __init__(self, bib_path: str, **kwargs):
        super().__init__(**kwargs)
        self._bib_path = bib_path
        self._entries: list[BibEntry] = []
        self._dirty = False
        self._first_run = is_first_run()
        self._config: Config = load_config()

    def compose(self) -> ComposeResult:
        yield Header()
        with Horizontal(id="main-content"):
            yield EntryList([], id="entry-list")
            yield EntryDetail(id="entry-detail")
        yield Footer()

    def on_mount(self) -> None:
        self.theme = detect_theme()
        self.title = f"bibtui — {os.path.basename(self._bib_path)}"
        self._load_entries()
        self._start_update_check()
        if self._first_run:
            self.call_after_refresh(self._show_first_run)

    def _start_update_check(self) -> None:
        if not self._config.check_for_updates:
            return
        self._check_for_updates()

    @work(thread=True)
    def _check_for_updates(self) -> None:
        now = update_check.utc_now()
        now_iso = update_check.to_utc_iso(now)

        if not update_check.is_due(self._config.update_last_check_utc, now):
            cached = self._config.update_latest_version
            if (
                cached
                and update_check.is_newer_version(__version__, cached)
                and not update_check.notified_today(
                    self._config.update_last_notified_utc, now
                )
            ):
                self.app.call_from_thread(
                    self._on_cached_update_available, cached, now_iso
                )
            return

        latest = update_check.fetch_latest_stable_version(timeout=3)
        self.app.call_from_thread(self._on_update_check_done, latest, now_iso)

    def _on_update_check_done(self, latest: str | None, checked_at_utc: str) -> None:
        self._config.update_last_check_utc = checked_at_utc

        if latest:
            self._config.update_latest_version = latest

        should_notify = bool(
            latest
            and update_check.is_newer_version(__version__, latest)
            and not update_check.notified_today(
                self._config.update_last_notified_utc,
                update_check.parse_utc_iso(checked_at_utc) or update_check.utc_now(),
            )
        )
        if should_notify:
            self._config.update_last_notified_utc = checked_at_utc

        save_config(self._config)

        if should_notify and latest:
            self._show_update_notification(latest)

    def _on_cached_update_available(self, latest: str, notified_at_utc: str) -> None:
        self._config.update_last_notified_utc = notified_at_utc
        save_config(self._config)
        self._show_update_notification(latest)

    def _show_update_notification(self, latest: str) -> None:
        self.notify(
            f"Update available: {latest} (installed {__version__}). "
            "Upgrade with: uv tool upgrade bibtui",
            timeout=8,
        )

    def on_resize(self, event) -> None:
        self.query_one("#main-content").set_class(
            event.size.width < event.size.height * 2, "vertical"
        )

    def on_paste(self, event: events.Paste) -> None:
        """Forward paste to a focused Input/TextArea, or open PasteModal for BibTeX text."""
        focused = self.focused
        if isinstance(focused, Input):
            focused.insert_text_at_cursor(event.text)
            return
        if isinstance(focused, TextArea):
            focused.insert(event.text)
            return
        # No text widget focused — intercept BibTeX-shaped pastes
        if event.text.strip().startswith("@"):
            event.stop()
            self.push_screen(PasteModal(event.text.strip()), self._on_paste_done)

    def _load_entries(self) -> None:
        self.notify("Loading bibliography...", timeout=2)
        try:
            import shutil as _shutil

            bak = self._bib_path + ".bak"
            if not os.path.exists(bak):
                _shutil.copy2(self._bib_path, bak)
        except Exception:
            pass  # Missing file or permission error — silently skip backup
        try:
            self._entries = parser.load(self._bib_path)
            entry_list = self.query_one(EntryList)
            entry_list.set_pdf_base_dir(self._config.pdf_base_dir)
            self.query_one(EntryDetail).set_pdf_base_dir(self._config.pdf_base_dir)
            entry_list.refresh_entries(self._entries)
            self.notify(f"Loaded {len(self._entries)} entries.", timeout=3)
        except Exception as e:
            self.notify(f"Error loading file: {e}", severity="error")
        self.query_one(DataTable).focus()

    # ── Entry selection ────────────────────────────────────────────────────

    @on(DataTable.RowHighlighted)
    def on_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        entry = self.query_one(EntryList).selected_entry
        self.query_one(EntryDetail).show_entry(entry)

    @on(DataTable.RowSelected)
    def on_row_selected(self, event: DataTable.RowSelected) -> None:
        entry = self.query_one(EntryList).selected_entry
        self.query_one(EntryDetail).show_entry(entry)

    # ── Actions ───────────────────────────────────────────────────────────

    async def action_quit(self) -> None:
        if not self._dirty:
            self.exit()
            return
        self.push_screen(
            ConfirmModal("You have unwritten changes. Quit without writing?"),
            lambda confirmed: self.exit() if confirmed else None,
        )

    def action_focus_search(self) -> None:
        self.query_one("#search-input", Input).focus()

    def action_clear_search(self) -> None:
        search = self.query_one("#search-input", Input)
        if search.value:
            search.value = ""
            search.blur()
        else:
            search.blur()

    def action_save(self) -> None:
        try:
            parser.save(self._entries, self._bib_path)
            self._dirty = False
            self.notify("Written.", timeout=3)
        except Exception as e:
            self.notify(f"Write failed: {e}", severity="error")

    def action_edit_entry(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        if self.query_one(EntryDetail).raw_mode:
            self.push_screen(RawEditModal(entry), self._on_edit_done)
        else:
            self.push_screen(EditModal(entry), self._on_edit_done)

    def _on_edit_done(self, result: BibEntry | None) -> None:
        if result is None:
            return
        # Update in master list
        for i, e in enumerate(self._entries):
            if e.key == result.key:
                self._entries[i] = result
                break
        self._dirty = True
        entry_list = self.query_one(EntryList)
        entry_list.refresh_entries(self._entries)
        self.query_one(EntryDetail).show_entry(result)
        self.notify("Entry updated. Press [w] to write.", timeout=3)

    def action_paste_import(self) -> None:
        self.push_screen(PasteModal(), self._on_paste_done)

    def _on_paste_done(self, result: BibEntry | None) -> None:
        if result is None:
            return
        existing_keys = {e.key for e in self._entries}
        if result.key in existing_keys:
            self.notify(
                f"BibTeX key '{result.key}' already exists, cannot proceed.",
                severity="error",
                timeout=5,
            )
            return
        self._entries.append(result)
        self._dirty = True
        el = self.query_one(EntryList)
        el.refresh_entries(self._entries)
        self.call_after_refresh(self._jump_to_entry, result)
        self.notify(f"Added: {result.key}", timeout=3)
        self._maybe_auto_fetch(result)

    def action_doi_import(self) -> None:
        self.push_screen(DOIModal(), self._on_doi_done)

    def action_delete_entry(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        self.push_screen(
            ConfirmModal(
                f"Delete entry [bold]{entry.key}[/bold]?\nThis cannot be undone."
            ),
            lambda confirmed: self._do_delete_entry(entry.key) if confirmed else None,
        )

    def _do_delete_entry(self, key: str) -> None:
        self._entries = [e for e in self._entries if e.key != key]
        self._dirty = True
        el = self.query_one(EntryList)
        el.refresh_entries(self._entries)
        self.query_one(EntryDetail).show_entry(el.selected_entry)
        self.notify(f"Deleted: {key}", timeout=3)

    def _on_doi_done(self, result: BibEntry | None) -> None:
        if result is None:
            return
        existing_keys = {e.key for e in self._entries}
        if result.key in existing_keys:
            self.notify(
                f"BibTeX key '{result.key}' already exists, cannot proceed.",
                severity="error",
                timeout=5,
            )
            return
        self._entries.append(result)
        self._dirty = True
        el = self.query_one(EntryList)
        el.refresh_entries(self._entries)
        self.call_after_refresh(self._jump_to_entry, result)
        self.notify(f"Added: {result.key}", timeout=3)
        self._maybe_auto_fetch(result)

    def _jump_to_entry(self, result: BibEntry) -> None:
        """Move cursor to the given entry after the table has been rendered."""
        el = self.query_one(EntryList)
        try:
            idx = next(i for i, e in enumerate(el._filtered) if e.key == result.key)
            self.query_one(DataTable).move_cursor(row=idx)
            self.query_one(EntryDetail).show_entry(result)
        except StopIteration:
            pass

    def _maybe_auto_fetch(self, entry: BibEntry) -> None:
        """Trigger PDF fetch after import if the setting is enabled and prerequisites are met."""
        if not self._config.auto_fetch_pdf:
            return
        if not (entry.doi or entry.url):
            return
        if not self._config.pdf_base_dir:
            return
        self._do_fetch_pdf(entry, True)

    def action_open_url(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        url = entry.url
        if not url:
            self.notify("No URL for this entry.", severity="warning")
            return
        parsed = urlparse(url)
        if parsed.scheme not in ("http", "https") or not parsed.netloc:
            self.notify(f"Invalid URL: {url}", severity="error", timeout=5)
            return
        webbrowser.open(url)
        self.notify(f"Opening: {url[:60]}", timeout=3)

    def action_open_openalex(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return

        title = entry.title.strip()
        doi = entry.doi.strip()
        query = title if title else doi
        if not query:
            self.notify("Entry has neither DOI nor title.", severity="warning")
            return

        openalex_url = f"https://openalex.org/works?search={quote_plus(query)}"
        webbrowser.open(openalex_url)

        if title:
            self.notify("Opening OpenAlex (title search)", timeout=3)
        else:
            self.notify("Opening OpenAlex (DOI search)", timeout=3)

    def action_fetch_pdf(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        if not entry.doi and not entry.url:
            self.notify(
                "Entry has no DOI or URL — cannot fetch PDF.",
                severity="warning",
            )
            return
        dest_dir = self._config.pdf_base_dir
        if not dest_dir:
            self.notify(
                "PDF base directory not set. Open Settings (Ctrl+P → Settings).",
                severity="warning",
            )
            return
        dest_path = os.path.join(dest_dir, pdf_filename(entry))
        has_valid_link = bool(
            entry.file and os.path.exists(parse_jabref_path(entry.file, dest_dir))
        )
        if os.path.exists(dest_path):
            if not has_valid_link:
                self._link_pdf_to_entry(
                    entry,
                    dest_path,
                    f"Found existing PDF and linked: {entry.key}",
                )
                return
            self.push_screen(
                ConfirmModal(f"PDF already exists:\n{dest_path}\n\nOverwrite?"),
                lambda confirmed: self._do_fetch_pdf(entry, confirmed),
            )
        else:
            self._do_fetch_pdf(entry, True)

    def _do_fetch_pdf(self, entry: BibEntry, confirmed: bool | None) -> None:
        if not confirmed:
            return
        self.push_screen(
            FetchPDFModal(
                entry,
                self._config.pdf_base_dir,
                self._config.unpaywall_email,
                overwrite=True,
            ),
            self._on_fetch_pdf_done,
        )

    def _on_fetch_pdf_done(self, result: str | None) -> None:
        if result is None:
            return
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        self._link_pdf_to_entry(entry, result, f"PDF saved and linked: {entry.key}")

    def action_add_pdf(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        dest_dir = self._config.pdf_base_dir
        if not dest_dir:
            self.notify(
                "PDF base directory not set. Open Settings (Ctrl+P → Settings).",
                severity="warning",
            )
            return
        self.push_screen(
            AddPDFModal(entry, dest_dir, self._config.pdf_download_dir),
            self._on_add_pdf_done,
        )

    def _on_add_pdf_done(self, result: str | None) -> None:
        if result is None:
            return
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        self._link_pdf_to_entry(entry, result, f"PDF added and linked: {entry.key}")

    def _link_pdf_to_entry(self, entry: BibEntry, path: str, message: str) -> None:
        entry.file = format_jabref_path(path, self._config.pdf_base_dir)
        self._dirty = True
        self.query_one(EntryList).refresh_row(entry)
        self.query_one(EntryDetail).show_entry(entry)
        self.notify(message, timeout=4)

    def action_open_pdf(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        if not entry.file:
            self.notify("No PDF linked for this entry.", severity="warning")
            return
        path = find_pdf_for_entry(entry.file, entry.key, self._config.pdf_base_dir)
        if not path:
            stored = parse_jabref_path(entry.file, self._config.pdf_base_dir)
            self.notify(f"PDF not found: {stored}", severity="error", timeout=5)
            return
        try:
            if platform.system() == "Darwin":
                subprocess.Popen(["open", path])
            else:
                subprocess.Popen(["xdg-open", path])
        except Exception as e:
            self.notify(f"Could not open PDF: {e}", severity="error", timeout=5)

    def action_set_rating(self, value: str) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        try:
            rating = max(0, min(5, int(value)))
        except ValueError:
            return
        entry.rating = rating
        self._dirty = True
        self.query_one(EntryList).refresh_row(entry)
        self.query_one(EntryDetail).show_entry(entry)

    def action_cycle_read_state(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        entry.cycle_read_state()
        self._dirty = True
        self.query_one(EntryList).refresh_row(entry)
        self.query_one(EntryDetail).show_entry(entry)

    def action_cycle_priority(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        entry.cycle_priority()
        self._dirty = True
        self.query_one(EntryList).refresh_row(entry)
        self.query_one(EntryDetail).show_entry(entry)

    def action_settings(self) -> None:
        self.push_screen(SettingsModal(self._config), self._on_settings_done)

    def action_fetch_missing_pdfs(self) -> None:
        self._start_library_fetch_missing_pdfs()

    def action_unify_citekeys(self) -> None:
        self._start_library_unify_citekeys()

    def _start_library_fetch_missing_pdfs(self) -> None:
        dest_dir = self._config.pdf_base_dir
        if not dest_dir:
            self.notify(
                "PDF base directory not set. Open Settings (Ctrl+P → Settings).",
                severity="warning",
            )
            return

        linked_count = self._autolink_existing_local_pdfs()
        if linked_count:
            self.notify(
                f"Auto-linked {linked_count} existing PDF"
                f"{'s' if linked_count != 1 else ''}.",
                timeout=4,
            )

        self.push_screen(LibraryFetchConfirmModal(), self._on_library_fetch_confirmed)

    def _on_library_fetch_confirmed(self, result: tuple[bool, bool] | None) -> None:
        if not result:
            return
        confirmed, overwrite_broken_links = result
        if not confirmed:
            return

        candidates = self._missing_pdf_candidates(overwrite_broken_links)
        if not candidates:
            self.notify("No missing PDFs found in this library.", timeout=4)
            return

        self._run_batch_fetch_missing_pdfs(candidates, overwrite_broken_links)

    def _missing_pdf_candidates(self, overwrite_broken_links: bool) -> list[BibEntry]:
        candidates: list[BibEntry] = []
        base_dir = self._config.pdf_base_dir
        for entry in self._entries:
            has_local_pdf = bool(find_pdf_for_entry(entry.file, entry.key, base_dir))
            if has_local_pdf:
                continue
            if entry.file and not overwrite_broken_links:
                continue
            candidates.append(entry)
        return candidates

    def _autolink_existing_local_pdfs(self) -> int:
        base_dir = self._config.pdf_base_dir
        linked = 0
        for entry in self._entries:
            stored_valid = bool(
                entry.file and os.path.exists(parse_jabref_path(entry.file, base_dir))
            )
            if stored_valid:
                continue

            found = find_pdf_for_entry(entry.file, entry.key, base_dir)
            if not found:
                continue

            entry.file = format_jabref_path(found, base_dir)
            linked += 1

        if linked:
            self._dirty = True
            self.query_one(EntryList).refresh_entries(self._entries)
            selected = self.query_one(EntryList).selected_entry
            self.query_one(EntryDetail).show_entry(selected)

        return linked

    def _run_batch_fetch_missing_pdfs(
        self, candidates: list[BibEntry], overwrite_broken_links: bool
    ) -> None:
        self.push_screen(
            BatchFetchPDFModal(
                candidates,
                self._config.pdf_base_dir,
                self._config.unpaywall_email,
                overwrite_broken_links=overwrite_broken_links,
            ),
            self._on_batch_fetch_missing_pdfs_done,
        )

    def _on_batch_fetch_missing_pdfs_done(self, result: dict | None) -> None:
        if not result:
            return

        paths_by_key: dict[str, str] = result.get("paths_by_key", {})
        if paths_by_key:
            for entry in self._entries:
                path = paths_by_key.get(entry.key)
                if path:
                    entry.file = format_jabref_path(path, self._config.pdf_base_dir)
            self._dirty = True
            self.query_one(EntryList).refresh_entries(self._entries)
            selected = self.query_one(EntryList).selected_entry
            self.query_one(EntryDetail).show_entry(selected)

        self.notify(
            f"Library fetch finished: {result.get('success', 0)} fetched, "
            f"{result.get('failed', 0)} failed, {result.get('skipped', 0)} skipped.",
            timeout=5,
        )

    def _start_library_unify_citekeys(self) -> None:
        scan = self._scan_citekey_unification()
        rename_count = len(scan["plan"])
        already_ok = scan["already_ok"]
        skipped_missing_metadata = scan["skipped_missing_metadata"]
        if rename_count == 0:
            self.notify(
                "No citekeys to change "
                f"({already_ok} already match AuthorYear"
                + (
                    f", {skipped_missing_metadata} skipped: missing author/year"
                    if skipped_missing_metadata
                    else ""
                )
                + ").",
                timeout=5,
            )
            return

        preview_count = min(3, rename_count)
        preview_lines = "\n".join(
            f"  {entry.key} → {new_key}"
            for entry, new_key in scan["plan"][:preview_count]
        )
        preview_more = (
            f"\n  … and {rename_count - preview_count} more"
            if rename_count > preview_count
            else ""
        )

        msg = (
            "This will change citation keys across the whole library.\n\n"
            "[bold]Warning:[/bold] Existing LaTeX documents using old citekeys may break.\n\n"
            f"Entries scanned: {scan['total']}\n"
            f"Already matching AuthorYear: {already_ok}\n"
            f"Will be renamed: {rename_count}\n\n"
            + (
                f"Skipped (missing author/year): {skipped_missing_metadata}\n\n"
                if skipped_missing_metadata
                else ""
            )
            + "Preview:\n"
            + preview_lines
            + preview_more
            + "\n\n"
            + "Continue?"
        )
        self.push_screen(
            ConfirmModal(msg),
            lambda confirmed: (
                self._apply_citekey_unification(scan["plan"]) if confirmed else None
            ),
        )

    def _scan_citekey_unification(self) -> dict:
        skipped_missing_metadata = 0
        used_keys: set[str] = set()
        plan: list[tuple[BibEntry, str]] = []
        already_ok = 0
        pending: list[BibEntry] = []

        for entry in self._entries:
            current_key = (entry.key or "").strip()
            if is_canonical_author_year_key(current_key):
                already_ok += 1
                used_keys.add(current_key)
                continue

            if not self._has_author_and_year(entry):
                skipped_missing_metadata += 1
                used_keys.add(current_key)
                continue

            pending.append(entry)

        for entry in pending:
            base = author_year_base(entry.author, entry.year)
            new_key = make_unique_key(base, used_keys)
            used_keys.add(new_key)
            if new_key == (entry.key or "").strip():
                already_ok += 1
            else:
                plan.append((entry, new_key))

        return {
            "total": len(self._entries),
            "already_ok": already_ok,
            "skipped_missing_metadata": skipped_missing_metadata,
            "plan": plan,
        }

    def _has_author_and_year(self, entry: BibEntry) -> bool:
        if not (entry.author or "").strip():
            return False
        return bool(re.search(r"\d{4}", entry.year or ""))

    def _apply_citekey_unification(self, plan: list[tuple[BibEntry, str]]) -> None:
        if not plan:
            return

        file_renamed = 0
        file_conflicts = 0
        base_dir = self._config.pdf_base_dir

        for entry, new_key in plan:
            old_key = entry.key

            if base_dir and entry.file:
                current_path = find_pdf_for_entry(entry.file, old_key, base_dir)
                if current_path and os.path.exists(current_path):
                    target_name = pdf_filename(
                        BibEntry(
                            key=new_key,
                            entry_type=entry.entry_type,
                            title=entry.title,
                        )
                    )
                    target_path = os.path.join(base_dir, target_name)
                    if os.path.abspath(current_path) != os.path.abspath(target_path):
                        if os.path.exists(target_path):
                            file_conflicts += 1
                        else:
                            os.replace(current_path, target_path)
                            entry.file = format_jabref_path(target_path, base_dir)
                            file_renamed += 1

            entry.key = new_key

        self._dirty = True
        self.query_one(EntryList).refresh_entries(self._entries)
        selected = self.query_one(EntryList).selected_entry
        self.query_one(EntryDetail).show_entry(selected)

        self.notify(
            f"Citekeys unified: {len(plan)} renamed, {file_renamed} PDF files renamed"
            + (
                f", {file_conflicts} file renames skipped (target exists)."
                if file_conflicts
                else "."
            ),
            timeout=6,
        )

    def _show_first_run(self) -> None:
        def _after_welcome(open_settings: bool | None) -> None:
            # Always persist an empty config so this modal never appears again.
            if not CONFIG_PATH.exists():
                save_config(self._config)
            if open_settings:
                self.push_screen(SettingsModal(self._config), self._on_settings_done)

        self.push_screen(FirstRunModal(), _after_welcome)

    def _on_settings_done(self, result: Config | None) -> None:
        if result is None:
            return
        self._config = result
        save_config(result)
        self.query_one(EntryList).set_pdf_base_dir(result.pdf_base_dir)
        self.query_one(EntryDetail).set_pdf_base_dir(result.pdf_base_dir)
        self.query_one(EntryList).refresh_entries(self._entries)
        self.notify("Settings written.", timeout=2)

    def action_toggle_view(self) -> None:
        self.query_one(EntryDetail).toggle_view()

    def action_toggle_table_maximize(self) -> None:
        entry_list = self.query_one(EntryList)
        screen = self.screen
        if screen.maximized is entry_list:
            screen.minimize()
            return
        if screen.maximized is not None:
            screen.minimize()
        self.query_one(DataTable).focus()
        if not screen.maximize(entry_list, container=False):
            self.notify(
                "Cannot maximize table pane in current state.", severity="warning"
            )

    def action_show_help(self) -> None:
        self.push_screen(HelpModal())

    def action_copy_key(self) -> None:
        focused = self.focused
        if isinstance(focused, (Input, TextArea)):
            copy_action = getattr(focused, "action_copy", None)
            if callable(copy_action):
                copy_action()
                return

        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        self.copy_to_clipboard(entry.key)
        self.notify(f"Copied: {entry.key}", timeout=2)

    def action_edit_keywords(self) -> None:
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            self.notify("No entry selected.", severity="warning")
            return
        all_kws, kw_counts = self._all_keywords()
        self.push_screen(
            KeywordsModal(entry, all_kws, kw_counts), self._on_keywords_done
        )

    def _all_keywords(self) -> tuple[list[str], dict[str, int]]:
        """All unique keywords across the bib file, sorted by frequency descending."""
        from collections import Counter

        counter: Counter[str] = Counter()
        for e in self._entries:
            for kw in e.keywords_list:
                counter[kw] += 1
        return [kw for kw, _ in counter.most_common()], dict(counter)

    def _on_keywords_done(self, result: tuple[str, set[str]] | None) -> None:
        if result is None:
            return
        keywords_str, delete_everywhere = result
        entry = self.query_one(EntryList).selected_entry
        if entry is None:
            return
        entry.keywords = keywords_str
        if delete_everywhere:
            for e in self._entries:
                if e is not entry:
                    kept = [k for k in e.keywords_list if k not in delete_everywhere]
                    e.keywords = ", ".join(kept)
        self._dirty = True
        self.query_one(EntryList).refresh_entries(self._entries)
        self.query_one(EntryDetail).show_entry(entry)
        self.notify("Keywords updated. Press [w] to write.", timeout=3)



import re
import unicodedata
from string import ascii_lowercase

_AUTHOR_YEAR_RE = re.compile(r"^[A-Z][A-Za-z0-9-]*\d{4}[a-z]?$")
_AUTHOR_YEAR_PARTS_RE = re.compile(r"^([A-Za-z0-9-]+?)(\d{4})([A-Za-z]?)$")
_YEAR_RE = re.compile(r"(\d{4})")


def is_author_year_key(key: str) -> bool:
    return bool(_AUTHOR_YEAR_RE.fullmatch((key or "").strip()))


def canonicalize_author_year_key(key: str) -> str:
    """Normalize casing for AuthorYear-like keys.

    Examples:
    - STEINIGER2021 -> Steiniger2021
    - steininger2021a -> Steininger2021a
    """
    k = (key or "").strip()
    match = _AUTHOR_YEAR_PARTS_RE.fullmatch(k)
    if not match:
        return k
    author_part, year_part, suffix = match.groups()
    if not author_part:
        return k
    author_part = "-".join(
        segment[0].upper() + segment[1:].lower() if segment else segment
        for segment in author_part.split("-")
    )
    return f"{author_part}{year_part}{suffix.lower()}"


def is_canonical_author_year_key(key: str) -> bool:
    k = (key or "").strip()
    return is_author_year_key(k) and canonicalize_author_year_key(k) == k


def author_year_base(author_field: str, year_field: str) -> str:
    surname = _extract_primary_surname(author_field)
    year = _extract_year(year_field)
    return f"{surname}{year}"


def make_unique_key(base_key: str, used: set[str]) -> str:
    if base_key not in used:
        return base_key
    for suffix in ascii_lowercase:
        candidate = f"{base_key}{suffix}"
        if candidate not in used:
            return candidate
    n = 2
    while True:
        candidate = f"{base_key}z{n}"
        if candidate not in used:
            return candidate
        n += 1


def _extract_year(year_field: str) -> str:
    match = _YEAR_RE.search(year_field or "")
    if match:
        return match.group(1)
    return "0000"


def _extract_primary_surname(author_field: str) -> str:
    if not author_field:
        return "Unknown"

    first_author = (author_field.split(" and ")[0] or "").strip()
    if not first_author:
        return "Unknown"

    if "," in first_author:
        surname_raw = first_author.split(",", 1)[0]
    else:
        parts = first_author.split()
        surname_raw = parts[-1] if parts else first_author

    normalized = _normalize_token(surname_raw)
    if not normalized:
        return "Unknown"

    if normalized.isupper():
        normalized = normalized.title()

    return normalized[0].upper() + normalized[1:]


def _normalize_token(text: str) -> str:
    cleaned = _normalize_latex_text(text)
    for src, dst in (
        ("ä", "ae"),
        ("Ä", "Ae"),
        ("ö", "oe"),
        ("Ö", "Oe"),
        ("ü", "ue"),
        ("Ü", "Ue"),
        ("ß", "ss"),
    ):
        cleaned = cleaned.replace(src, dst)
    cleaned = unicodedata.normalize("NFKD", cleaned)
    cleaned = "".join(ch for ch in cleaned if not unicodedata.combining(ch))
    cleaned = re.sub(r"[^A-Za-z0-9]", "", cleaned)
    return cleaned


def _normalize_latex_text(text: str) -> str:
    t = text or ""
    for src, dst in (
        (r"\\ss", "ss"),
        (r"\\SS", "SS"),
        (r"\\ae", "ae"),
        (r"\\AE", "AE"),
        (r"\\oe", "oe"),
        (r"\\OE", "OE"),
        (r"\\aa", "aa"),
        (r"\\AA", "AA"),
        (r"\\o", "o"),
        (r"\\O", "O"),
        (r"\\l", "l"),
        (r"\\L", "L"),
    ):
        t = t.replace(src, dst)

    def _umlaut_repl(match: re.Match[str]) -> str:
        ch = match.group(1)
        return {
            "a": "ae",
            "A": "Ae",
            "o": "oe",
            "O": "Oe",
            "u": "ue",
            "U": "Ue",
        }.get(ch, ch)

    t = re.sub(r'\\"\s*\{?\s*([A-Za-z])\s*\}?', _umlaut_repl, t)

    # Common one-letter accent macros, e.g. {\"o}, \'e, \v{c}
    t = re.sub(
        r"\\[\"'`^~=.uvHckrbdt]\s*\{?\s*([A-Za-z])\s*\}?",
        r"\1",
        t,
    )

    # Remaining command wrappers around one token: \textit{X} -> X
    t = re.sub(r"\\[A-Za-z]+\s*\{([^}]*)\}", r"\1", t)

    # Remove any remaining commands and braces
    t = re.sub(r"\\[A-Za-z]+", "", t)
    t = t.replace("{", "").replace("}", "")

    return t



import click

from bibtui import __version__


@click.command()
@click.version_option(__version__, prog_name="bibtui")
@click.argument("bib_file", type=click.Path(exists=True, readable=True, dir_okay=False))
def main(bib_file: str) -> None:
    """Browse and manage BibTeX bibliography files.

    BIB_FILE is the path to the .bib file to open.

    \b
    Examples:
      bibtui references.bib
      bibtui ~/papers/MyCollection.bib

    \b
    Docs & source: https://github.com/tgoelles/bib_tui
    """
    from bibtui.app import BibTuiApp

    BibTuiApp(bib_file).run()


if __name__ == "__main__":
    main()
